import warnings

import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord, SkyOffsetFrame, ICRS
from astropy.time import Time

from sora.config.decorators import deprecated_alias
from .meta import BaseEphem

__all__ = ['EphemHorizons', 'EphemJPL', 'EphemKernel', 'EphemPlanete']


class EphemPlanete(BaseEphem):
    """Class used to simulate former Fortran programs `ephem_planete` and
    `fit_d2_ksi_eta`.

    Attributes
    ----------
    ephem : `file`, required
        Input file with JD (UTC), geocentric RA (deg), DEC (deg), and
        distance (AU).

    name : `str`, optional, default=None
        Name of the object to search in the JPL database.

    radius : `int`, `float`, optional, default: online database
        Object radius, in km.

    error_ra : `int`, `float`, optional, default: online database
        Ephemeris RA*cosDEC error, in arcsec.

    error_dec : `int`, `float`, optional, default: online database
        Ephemeris DEC error, in arcsec.

    mass : `int`, `float`, optional. default=0
        Object mass, in kg.

    H : `int`, `float`, optional, default=NaN
        Object absolute magnitude.

    G : `int`, `float`, optional, default=NaN
        Object phase slope.

    """

    def __init__(self, ephem, name=None, spkid=None, **kwargs):

        super().__init__(name=name, spkid=spkid, **kwargs)
        data = np.loadtxt(ephem, unpack=True)
        self.time = Time(data[0], format='jd')
        self.ephem = SkyCoord(data[1] * u.deg, data[2] * u.deg, data[3] * u.AU)
        self.__reftime = self.time[0]
        self.min_time = Time(data[0].min(), format='jd')
        self.max_time = Time(data[0].max(), format='jd')

    def get_position(self, time, observer='geocenter'):
        """Returns the object's geocentric position.

        Parameters
        ----------
        time : `str`, `astropy.time.Time`
            Reference time to calculate the object position. It can be a string
            in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.

        observer : `any`
            This parameter is present in EphemPlanete for compatibility with the
            remaining ephem classes. The returned positions are based on the given
            ephemeris despite the observer.

        Returns
        -------
        coord : `astropy.coordinates.SkyCoord`
            Astropy SkyCoord object with the object coordinates at the given time.
        """
        ksi, eta = self.get_ksi_eta(time=time) * u.km
        distance = self.ephem.distance.mean()
        off_ra = np.arctan2(ksi, distance)
        off_dec = np.arctan2(eta, distance)
        coord_frame = SkyOffsetFrame(origin=self.star)
        pos = SkyCoord(lon=off_ra, lat=off_dec, distance=distance, frame=coord_frame)
        return pos.icrs

    @deprecated_alias(log='verbose')  # remove this line in v1.0
    def fit_d2_ksi_eta(self, star, verbose=True):
        """Fits the projected position (orthographic projection) of the object in
        the tangent sky plane relative to a star.

        Parameters
        ----------
        star : `str`, `astropy.coordinates.SkyCoord`
            The coordinate of the star in the same reference frame as the ephemeris.

        verbose : `bool`, optional, default=True
            Enable log printing.
        """
        if type(star) == str:
            star = SkyCoord(star, unit=(u.hourangle, u.deg))
        if hasattr(self, 'star') and self.star.to_string('hmsdms', precision=5) == star.to_string('hmsdms',
                                                                                                  precision=5):
            return
        self.star = star
        target = self.ephem.transform_to(SkyOffsetFrame(origin=star))
        da = target.cartesian.y
        dd = target.cartesian.z
        dt = (self.time - self.__reftime) / (self.max_time - self.min_time)

        self.ksi = np.polyfit(dt, da.to(u.km).value, 2)
        self.eta = np.polyfit(dt, dd.to(u.km).value, 2)
        ksi = np.poly1d(self.ksi)
        eta = np.poly1d(self.eta)
        dksi = da.to(u.km).value - ksi(dt)
        deta = dd.to(u.km).value - eta(dt)
        rmsk = np.sqrt(np.mean(np.square(dksi)))
        rmse = np.sqrt(np.mean(np.square(deta)))
        if verbose:
            output = ('Fitting ephemeris position relative to star coordinate {}\n'
                      'ksi = aksi*t\u00b2 + bksi*t + cksi\n'
                      'eta = aeta*t\u00b2 + beta*t + ceta\n'
                      't=(jd-{})/({}-{})\n'
                      '        aksi={}\n'
                      '        bksi={}\n'
                      '        cksi={}\n'
                      '        aeta={}\n'
                      '        beta={}\n'
                      '        ceta={}\n'
                      'Residual RMS: ksi={:.3f} km, eta={:.3f} km'.format(
                          self.star.to_string('hmsdms'), self.__reftime.jd, self.max_time.jd,
                          self.min_time.jd, *self.ksi, *self.eta, rmsk, rmse)
                      )
            print(output)

    def get_ksi_eta(self, time, star=None):
        """Returns the projected position (orthographic projection) of the object
        in the tangent sky plane relative to a star.

        Parameters
        ----------
        time : `str`, `astropy.time.Time`, required
            Reference time to calculate the position. It can be a string
            in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.

        star : `str`, `astropy.coordinates.SkyCoord`, optional, default=None
            The coordinate of the star in the same reference frame as the ephemeris.

        Returns
        -------
        ksi, eta : `float` array
            Projected position (orthographic projection) of the object in the
            tangent sky plane relative to a star.
            ``ksi`` is in the North-South direction (North positive).
            ``eta`` is in the East-West direction (East positive).
        """
        if star:
            self.fit_d2_ksi_eta(star)
        time = Time(time)
        if not time.isscalar:
            if any(time < self.min_time) or any(time > self.max_time):
                raise ValueError('time must be in the interval [{},{}]'.format(
                    self.min_time, self.max_time))
        elif time.isscalar:
            if time < self.min_time or time > self.max_time:
                raise ValueError('time must be in the interval [{},{}]'.format(
                    self.min_time, self.max_time))
        if hasattr(self, 'ksi') and hasattr(self, 'eta'):
            ksi = np.poly1d(self.ksi)
            eta = np.poly1d(self.eta)
            k = ksi((time - self.__reftime) / (self.max_time - self.min_time))
            e = eta((time - self.__reftime) / (self.max_time - self.min_time))
            if hasattr(self, 'offset'):
                dist = self.ephem[0].distance.to(u.km).value
                dksi = np.sin(self.offset.d_lon_coslat) * dist
                deta = np.sin(self.offset.d_lat) * dist
                k = k + dksi
                e = e + deta
            return k.value, e.value
        else:
            raise ValueError('A "star" parameter is missing. Please run fit_d2_ksi_eta first.')

    def __str__(self):
        """ String representation of the EphemPlanete Class.
        """
        validity = 'Valid from {} until {}'.format(self.min_time.iso, self.max_time.iso)
        out = super().__str__().format(ephem_info=validity)
        if hasattr(self, 'star'):
            out += ("\nFitted ephemeris position relative to star coordinate {}\n"
                    "    ksi = aksi*t\u00b2 + bksi*t + cksi\n"
                    "    eta = aeta*t\u00b2 + beta*t + ceta\n"
                    "    t=(jd-{})/({}-{})\n"
                    "        aksi={}\n"
                    "        bksi={}\n"
                    "        aeta={}\n"
                    "        beta={}\n"
                    "        ceta={}\n".format(
                        self.star.to_string('hmsdms'), self.__reftime.jd, self.max_time.jd,
                        self.min_time.jd, *self.ksi, *self.eta)
                    )
            target = self.ephem.transform_to(SkyOffsetFrame(origin=self.star))
            da = -target.cartesian.y
            dd = -target.cartesian.z
            dt = (self.time - self.__reftime) / (self.max_time - self.min_time)
            ksi = np.poly1d(self.ksi)
            eta = np.poly1d(self.eta)
            dksi = da.to(u.km).value - ksi(dt)
            deta = dd.to(u.km).value - eta(dt)
            rmsk = np.sqrt(np.mean(np.square(dksi)))
            rmse = np.sqrt(np.mean(np.square(deta)))
            out += 'Residual RMS: ksi={:.3f} km, eta={:.3f} km\n'.format(rmsk, rmse)
        return out


class EphemHorizons(BaseEphem):
    """Obtains the ephemeris from Horizons/JPL service.

    Note
    ----
    Web tool URL: https://ssd.jpl.nasa.gov/horizons.cgi


    Attributes
    ----------
    name : `str`, required
        Name of the object to search in the JPL database.

    id_type: `str`, default='smallbody'
        Type of object options: ``smallbody``, ``majorbody`` (planets but
        also anything that is not a small body), ``designation``, ``name``,
        ``asteroid_name``, ``comet_name``, ``id`` (Horizons id number), or
        ``smallbody`` (find the closest match under any id_type).

    radius : `int`, `float`, default: online database
        Object radius, in km.

    error_ra : `int`, `float`, default: online database
        Ephemeris RA*cosDEC error, in arcsec.

    error_dec : `int`, `float`, default: online database
        Ephemeris DEC error, in arcsec.

    mass : `int`, `float`, default=0
        Object mass, in kg.

    H : `int`, `float`, default=NaN
        Object absolute magnitude.

    G : `int`, `float`, default=NaN
        Object phase slope.

    """

    def __init__(self, name, id_type='smallbody', spkid=None, **kwargs):

        super().__init__(name=name, spkid=spkid, **kwargs)
        self.id_type = id_type
        _ = self.get_position(Time.now())  # test if Horizons can proceed for this object

    def get_position(self, time, observer='geocenter'):
        """Returns the ICRS position of the object for observer.

        Parameters
        ----------
        time : `str`, `astropy.time.Time`
            Reference time to calculate the object position. It can be a string
            in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.

        observer : `str`, `sora.Observer`, `sora.Spacecraft`
            IAU code of the observer (must be present in given list of kernels),
            a SORA observer object or a string: ['geocenter', 'barycenter']

        Returns
        -------
        coord : `astropy.coordinates.SkyCoord`
            Astropy SkyCoord object with the object coordinates at the given time.
        """
        from .utils import ephem_horizons

        coord = ephem_horizons(time=time, target=self.name, observer=observer, id_type=self.id_type, output='ephemeris')
        if hasattr(self, 'offset'):
            pos_frame = SkyOffsetFrame(origin=coord)
            new_pos = SkyCoord(lon=self.offset.d_lon_coslat, lat=self.offset.d_lat,
                               distance=coord.distance, frame=pos_frame)
            coord = new_pos.transform_to(ICRS)
        if not coord.isscalar and len(coord) == 1:
            coord = coord[0]
        return coord

    def __str__(self):
        """ String representation of the EphemHorizons Class.
        """
        out = super().__str__().format(ephem_info='Ephemeris are downloaded from Horizons website')
        return out


# remove this block for v1.0
class EphemJPL(EphemHorizons):
    def __init__(self, name, id_type='smallbody', spkid=None, **kwargs):
        warnings.warn('EphemJPL is deprecated and will be removed in v1.0. Please use EphemHorizons.')
        super().__init__(name=name, id_type=id_type, spkid=spkid, **kwargs)

    __init__.__doc__ = EphemHorizons.__init__.__doc__
# end of block removal for v1.0


class EphemKernel(BaseEphem):
    """Gets the ephemeris from BSP kernels.

    Parameters
    ----------
    name : `str`,  optional, default=None
        Name of the object to search in the JPL database.

    spkid : `str`, required
        `spkid` of the targeting object. Former 'code' (v0.1).

    kernels : `list`, required
        List of paths for kernels files.

    radius : `int`, `float`, optional, default: online database
        Object radius, in km.

    error_ra : `int`, `float`, optional, default: online database
        Ephemeris RA*cosDEC error, in arcsec .

    error_dec : `int`, `float`, optional, default: online database
        Ephemeris DEC error, in arcsec.

    mass : `int`, `float`, optional, default=0
        Object Mass, in kg.

    H : `int`, `float`, optional, default=NaN
        Object Absolute Magnitude.

    G : `int`, `float`, optional, default=NaN
        Object Phase slope.

    """

    @deprecated_alias(code='spkid')  # remove this line for v1.0
    def __init__(self, kernels, spkid, name=None, **kwargs):
        import spiceypy as spice

        super().__init__(name=name, spkid=spkid, **kwargs)
        self.meta = {}
        kerns = []
        for arg in kernels:
            spice.furnsh(arg)
            kerns.append(arg.split('/')[-1].split('.')[0].upper())
        spice.kclear()
        self.meta['kernels'] = '/'.join(kerns)
        self.__kernels = kernels

    def get_position(self, time, observer='geocenter'):
        """Returns the object geocentric position.

        Parameters
        ----------
        time : `str`, `astropy.time.Time`
            Reference time to calculate the object position. It can be a string
            in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.

        observer : `str`, `sora.Observer`, `sora.Spacecraft`
            IAU code of the observer (must be present in given list of kernels),
            a SORA observer object or a string: ['geocenter', 'barycenter']

        Returns
        -------
        coord : `astropy.coordinates.SkyCoord`
            Astropy SkyCoord object with the object coordinates at the given time.
        """
        from .utils import ephem_kernel
        pos = ephem_kernel(time=time, target=self.spkid, observer=observer, kernels=self.__kernels)
        if hasattr(self, 'offset'):
            pos_frame = SkyOffsetFrame(origin=pos)
            new_pos = SkyCoord(lon=self.offset.d_lon_coslat, lat=self.offset.d_lat,
                               distance=pos.distance, frame=pos_frame)
            pos = new_pos.transform_to(ICRS)
        pos = SkyCoord(ra=pos.ra, dec=pos.dec, distance=pos.distance)
        if not pos.isscalar and len(pos) == 1:
            pos = pos[0]
        return pos

    def __str__(self):
        """ String representation of the EphemKernel Class.
        """
        out = super().__str__().format(ephem_info=self.meta['kernels'])
        return out



import astropy.units as u
import numpy as np

__all__ = ['search_code_mpc']


def search_code_mpc():
    """Reads the MPC Observer Database.

    Returns
    -------
    observatories : `dict`
        A python dictionary with all the sites as Astropy EarthLocation objects.
    """
    from astropy.coordinates import EarthLocation
    from astroquery.mpc import MPC

    obs = MPC.get_observatory_codes()
    observatories = {}
    for line in obs:
        code = line['Code']
        lon = line['Longitude'] * u.deg
        rcphi = line['cos'] * 6378.137 * u.km
        rsphi = line['sin'] * 6378.137 * u.km
        name = line['Name']
        site = EarthLocation.from_geocentric(rcphi * np.cos(lon), rcphi * np.sin(lon), rsphi)
        observatories[code] = (name, site)
    return observatories



import warnings

import astropy.constants as const
import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.time import Time

from sora.body import Body
from sora.ephem.meta import BaseEphem
from sora.star import Star
from sora.config.decorators import deprecated_alias

__all__ = ['occ_params', 'prediction']


def occ_params(star, ephem, time, n_recursions=5, max_tdiff=None, reference_center='geocenter'):
    """Calculates the parameters of the occultation, as instant, CA, PA.

    Parameters
    ----------
    star : `sora.Star`
        The coordinate of the star in the same reference frame as the ephemeris.
        It must be a Star object.

    ephem : `sora.Ephem*`
        Object ephemeris. It must be an Ephemeris object.

    time : `astropy.time.Time`
        Time close to occultation epoch to calculate occultation parameters.

    n_recursions : `int`, default=5
        The number of attempts to try obtain prediction parameters in case the
        event is outside the previous range of time.

    max_tdiff : `int`, default=None
        Maximum difference from given time it will attempt to identify the
        occultation, in minutes. If given, 'n_recursions' is ignored.

    reference_center : `str`, `sora.Observer`, `sora.Spacecraft`
            A SORA observer object or a string 'geocenter'.
            The occultation parameters will be calculated in respect
            to this reference as center of projection.

    Returns
    -------
     Oredered list : `list`
        - Instant of CA (Time): Instant of Closest Approach.\n
        - CA (arcsec): Distance of Closest Approach.\n
        - PA (deg): Position Angle at Closest Approach.\n
        - vel (km/s): Velocity of the occultation.\n
        - dist (AU): the object geocentric distance.\n
    """
    from sora.ephem import EphemPlanete
    from sora.observer import Observer, Spacecraft
    from astropy.coordinates import SkyOffsetFrame

    n_recursions = int(n_recursions)
    n_iter = n_recursions
    if type(star) != Star:
        raise ValueError('star must be a Star object')
    if not isinstance(ephem, BaseEphem):
        raise ValueError('ephem must be an Ephemeris object')
    if reference_center != 'geocenter' and not isinstance(reference_center, (Observer, Spacecraft)):
        raise ValueError('reference_center must be "geocenter" or an observer object.')

    time = Time(time)
    pos_star = star.get_position(time=time, observer=reference_center)

    if isinstance(ephem, EphemPlanete):
        ephem.fit_d2_ksi_eta(star=pos_star, verbose=False)

    def calc_min(time0, time_interval, delta_t, n_recursions=5, max_tdiff=None):
        if max_tdiff is not None:
            max_t = u.Quantity(max_tdiff, unit=u.min)
            if np.absolute((time0 - time).sec*u.s) > max_t - time_interval*u.s:
                raise ValueError('Occultation is farther than {} from given time'.format(max_t))
        if n_recursions == 0:
            raise ValueError('Occultation is farther than {} min from given time'.format(n_iter*time_interval/60))
        tt = time0 + np.arange(-time_interval, time_interval, delta_t)*u.s
        pos_ephem = ephem.get_position(time=tt, observer=reference_center)
        _, f, g, = - pos_ephem.transform_to(SkyOffsetFrame(origin=pos_star)).cartesian.xyz.to(u.km).value
        dd = np.sqrt(f*f+g*g)
        min = np.argmin(dd)
        if min < 2:
            return calc_min(time0=tt[0], time_interval=time_interval, delta_t=delta_t,
                            n_recursions=n_recursions-1, max_tdiff=max_tdiff)
        elif min > len(tt) - 2:
            return calc_min(time0=tt[-1], time_interval=time_interval, delta_t=delta_t,
                            n_recursions=n_recursions-1, max_tdiff=max_tdiff)

        return tt[min]

    tmin = calc_min(time0=time, time_interval=900, delta_t=20, n_recursions=4, max_tdiff=max_tdiff)
    tmin = calc_min(time0=tmin, time_interval=20, delta_t=0.5, n_recursions=4, max_tdiff=max_tdiff)
    tmin = calc_min(time0=tmin, time_interval=1, delta_t=0.02, n_recursions=5, max_tdiff=max_tdiff)

    pos_ephem = ephem.get_position(time=tmin, observer=reference_center)
    _, f, g, = - pos_ephem.transform_to(SkyOffsetFrame(origin=pos_star)).cartesian.xyz.to(u.km).value
    dd = np.sqrt(f * f + g * g)
    dist = pos_ephem.distance
    ca = np.arcsin(dd * u.km / dist).to(u.arcsec)
    pa = (np.arctan2(-f, -g) * u.rad).to(u.deg)
    if pa < 0 * u.deg:
        pa = pa + 360 * u.deg

    pos_ephem = ephem.get_position(time=tmin + 1 * u.s, observer=reference_center)
    _, f2, g2, = - pos_ephem.transform_to(SkyOffsetFrame(origin=pos_star)).cartesian.xyz.to(u.km).value
    df = f2 - f
    dg = g2 - g
    vel = np.sqrt(df ** 2 + dg ** 2) / 1
    vel = vel * np.sign(dg) * (u.km / u.s)

    return tmin, ca, pa, vel, dist.to(u.AU)


@deprecated_alias(log='verbose')  # remove this line in v1.0
def prediction(time_beg, time_end, body=None, ephem=None, mag_lim=None, catalogue='gaiaedr3', step=60, divs=1, sigma=1,
               radius=None, verbose=True, reference_center='geocenter'):
    """Predicts stellar occultations.

    Parameters
    ----------
    time_beg : `str`, `astropy.time.Time`, required
        Initial time for prediction.

    time_end : `str`, `astropy.time.Time`, required
        Final time for prediction.

    body : `sora.Body`, `str`, default=None
        Object that will occult the stars. It must be a Body object or its name
        to search in the Small Body Database.

    ephem : `sora.Ephem`, default=None
        object ephemeris. It must be an Ephemeris object.
        If using a EphemHorizons object, please use 'divs' to make division
        at most a month, or a timeout error may be raised by the Horizon query.

    mag_lim : `int`, `float`, default=None
        Faintest Gmag allowed in the search.

    catalogue : `str`, default='gaiaedr3'
        The catalogue to download data. It can be ``'gaiadr2'`` or ``'gaiaedr3'``.

    step : `int`, `float`, default=60
        Step, in seconds, of ephem times for search

    divs : `int`, default=1
        Number of regions the ephemeris will be split for better search of
        occultations.

    sigma : `int`, `float`, default=1
        Ephemeris error sigma for search off-Earth.

    radius : `int`, `float`, default=None
        The radius of the body. It is important if not defined in body or ephem.

    verbose : `bool`, default=True
        To show what is being done at the moment.

    reference_center : `str`, `sora.Observer`, `sora.Spacecraft`
        A SORA observer object or a string 'geocenter'.
        The occultation parameters will be calculated in respect
        to this reference as center of projection. If a Spacecraft
        is used, please use smaller step since the search will be based
        on the target size and ephemeris error only.


    Important
    ---------
    When instantiating with "body" and "ephem", the user may call the function
    in 3 ways:

    1 - With "body" and "ephem".

    2 - With only "body". In this case, the "body" parameter must be a Body
    object and have an ephemeris associated (see Body documentation).

    3 - With only "ephem". In this case, the "ephem" parameter must be one of
    the Ephem Classes and have a name (see Ephem documentation) to search
    for the body in the Small Body Database.

    Returns
    -------
     : `sora.prediction.PredictionTable`
        PredictionTable with the occultation params for each event.
    """
    from astroquery.vizier import Vizier
    from sora.observer import Observer, Spacecraft
    from .table import PredictionTable

    if reference_center != 'geocenter' and not isinstance(reference_center, (Observer, Spacecraft)):
        raise ValueError('reference_center must be "geocenter" or an observer object.')

    # generate ephemeris
    if body is None and ephem is None:
        raise ValueError('"body" and/or "ephem" must be given.')
    if body is not None:
        if not isinstance(body, (str, Body)):
            raise ValueError('"body" must be a string with the name of the object or a Body object')
        if isinstance(body, str):
            body = Body(name=body)
    if ephem is not None:
        if body is not None:
            body.ephem = ephem
            ephem = body.ephem
    else:
        ephem = body.ephem
    time_beg = Time(time_beg)
    time_end = Time(time_end)
    intervals = np.round(np.linspace(0, (time_end-time_beg).sec, divs+1))

    # define catalogue parameters
    allowed_catalogues = ['gaiadr2', 'gaiaedr3']
    if catalogue not in allowed_catalogues:
        raise ValueError('Catalogue {} is not one of the allowed catalogues {}'.format(catalogue, allowed_catalogues))
    vizpath = {'gaiadr2': 'I/345/gaia2', 'gaiaedr3': 'I/350/gaiaedr3'}[catalogue]
    cat_name = {'gaiadr2': 'Gaia-DR2', 'gaiaedr3': 'Gaia-EDR3'}[catalogue]
    columns = {'gaiadr2': ['Source', 'RA_ICRS', 'DE_ICRS', 'pmRA', 'pmDE', 'Plx', 'RV', 'Epoch', 'Gmag'],
               'gaiaedr3': ['Source', 'RA_ICRS', 'DE_ICRS', 'pmRA', 'pmDE', 'Plx', 'RVDR2', 'Epoch', 'Gmag']}[catalogue]
    kwds = {'columns': columns, 'row_limit': 10000000, 'timeout': 600}
    if mag_lim:
        kwds['column_filters'] = {"Gmag": "<{}".format(mag_lim)}
    vquery = Vizier(**kwds)

    # determine suitable divisions for star search
    if radius is None:
        try:
            radius = ephem.radius  # for v1.0, change to body.radius
        except AttributeError:
            radius = 0
            warnings.warn('"radius" not given or found in body or ephem. Considering it to be zero.')
        if np.isnan(radius):
            radius = 0
    radius = u.Quantity(radius, unit=u.km)

    radius_search = radius
    if reference_center == 'geocenter':
        radius_search = radius_search + const.R_earth

    if verbose:
        print('Ephemeris was split in {} parts for better search of stars'.format(divs))

    # makes predictions for each division
    occs = []
    for i in range(divs):
        dt = np.arange(intervals[i], intervals[i+1], step)*u.s
        nt = time_beg + dt
        if verbose:
            print('\nSearching occultations in part {}/{}'.format(i+1, divs))
            print("Generating Ephemeris between {} and {} ...".format(nt.min(), nt.max()))
        ncoord = ephem.get_position(time=nt, observer=reference_center)
        ra = np.mean([ncoord.ra.min().deg, ncoord.ra.max().deg])
        dec = np.mean([ncoord.dec.min().deg, ncoord.dec.max().deg])
        mindist = (np.arcsin(radius_search/ncoord.distance).max() +
                   sigma*np.max([ephem.error_ra.value, ephem.error_dec.value])*u.arcsec)
        width = ncoord.ra.max() - ncoord.ra.min() + 2*mindist
        height = ncoord.dec.max() - ncoord.dec.min() + 2*mindist
        pos_search = SkyCoord(ra*u.deg, dec*u.deg)

        if verbose:
            print('Downloading stars ...')
        catalogue = vquery.query_region(pos_search, width=width, height=height, catalog=vizpath, cache=False)
        if len(catalogue) == 0:
            print('    No star found. The region is too small or VizieR is out.')
            continue
        catalogue = catalogue[0]
        if verbose:
            print('    {} {} stars downloaded'.format(len(catalogue), cat_name))
            print('Identifying occultations ...')
        pm_ra_cosdec = catalogue['pmRA'].quantity
        pm_ra_cosdec[np.where(np.isnan(pm_ra_cosdec))] = 0*u.mas/u.year
        pm_dec = catalogue['pmDE'].quantity
        pm_dec[np.where(np.isnan(pm_dec))] = 0*u.mas/u.year
        stars = SkyCoord(catalogue['RA_ICRS'].quantity, catalogue['DE_ICRS'].quantity, distance=np.ones(len(catalogue))*u.pc,
                         pm_ra_cosdec=pm_ra_cosdec, pm_dec=pm_dec,
                         obstime=Time(catalogue['Epoch'].quantity.value, format='jyear'))
        prec_stars = stars.apply_space_motion(new_obstime=((nt[-1]-nt[0])/2+nt[0]))
        idx, d2d, d3d = prec_stars.match_to_catalog_sky(ncoord)

        dist = np.arcsin(radius_search/ncoord[idx].distance) + sigma*np.max([ephem.error_ra.value,
                                                                             ephem.error_dec.value])*u.arcsec \
            + np.sqrt(stars.pm_ra_cosdec**2+stars.pm_dec**2)*(nt[-1]-nt[0])/2
        k = np.where(d2d < dist)[0]
        for ev in k:
            star = Star(code=catalogue['Source'][ev], ra=catalogue['RA_ICRS'][ev]*u.deg, dec=catalogue['DE_ICRS'][ev]*u.deg,
                        pmra=catalogue['pmRA'][ev]*u.mas/u.year, pmdec=catalogue['pmDE'][ev]*u.mas/u.year,
                        parallax=catalogue['Plx'][ev]*u.mas, rad_vel=catalogue[columns[6]][ev]*u.km/u.s,
                        epoch=Time(catalogue['Epoch'][ev], format='jyear'), local=True, nomad=False, verbose=False)
            c = star.get_position(time=nt[idx][ev], observer=reference_center)
            pars = [star.code, SkyCoord(c.ra, c.dec), catalogue['Gmag'][ev]]
            try:
                pars = np.hstack((pars, occ_params(star, ephem, nt[idx][ev], reference_center=reference_center)))
                occs.append(pars)
            except:
                pass

    meta = {'name': ephem.name or getattr(body, 'shortname', ''), 'time_beg': time_beg, 'time_end': time_end,
            'maglim': mag_lim, 'max_ca': mindist, 'radius': radius.to(u.km).value,
            'error_ra': ephem.error_ra.to(u.mas).value, 'error_dec': ephem.error_dec.to(u.mas).value,
            'ephem': ephem.meta['kernels'], 'catalogue': cat_name}
    if not occs:
        print('\nNo stellar occultation was found.')
        return PredictionTable(meta=meta)
    # create astropy table with the params
    occs2 = np.transpose(occs)
    time = Time(occs2[3])
    obj_pos = SkyCoord([ephem.get_position(time=time, observer=reference_center)])
    k = np.argsort(time)
    t = PredictionTable(
        time=time[k], coord_star=occs2[1][k], coord_obj=obj_pos[k], ca=[i.value for i in occs2[4][k]],
        pa=[i.value for i in occs2[5][k]], vel=[i.value for i in occs2[6][k]], mag=occs2[2][k],
        dist=[i.value for i in occs2[7][k]], source=occs2[0][k], meta=meta)
    if verbose:
        print('\n{} occultations found.'.format(len(t)))
    return t



praia_occ_head = (
    " Notes:\n"
    " C/A: geocentric closest approach, in arcsec\n"
    " P/A: Planet position angle wrt to star at C/A, in deg\n"
    " vel: velocity in plane of sky, in km/sec, positive= prograde, negative= retrograde\n"
    " R: instrumental magnitude in UCAC2 system\n"
    " J, H, K: 2MASS magnitudes (50.0 = not in 2MASS)\n"
    " G*, J*, H*, K* are normalized magnitudes to a common\n"
    " shadow velocity of 20 km/sec by the relationship:\n"
    " Mag* = Mag_actual + 2.5*log10[velocity/20 (km/sec)]\n"
    " Delta: Planet range to Earth, AU\n"
    " long: East longitude of sub-planet point, deg, positive towards East\n"
    " loc. t.= UT + long: local solar time at sub-planet point, hh:mm\n"
    " ----------------------------------------------------------------------------------"
    "----------------------------------\n"
    " Selection criteria:\n"
    " Maximum geocentric closest approach considered:   {max_ca:.3f}\n"
    " Day light exclusion (range loc. t.): 12.0hs to 12.0 hs (t1 = t2 -> no exclusions)\n"
    " List below has:           {size}  entries\n"
    " Reference ephemeris: {ephem}\n"
    " Offset applied to ephemeris off_ra(mas) = A * (t-t0) + B \n"
    " Offset applied to ephemeris off_de(mas) = C * (t-t0) + D \n"
    " t0 =    2005.0000000000000       yrs\n"
    " A =    0.0000000000000000       (mas/yr)\n"
    " B =    0.0000000000000000       (mas)\n"
    " C =    0.0000000000000000       (mas/yr)\n"
    " D =    0.0000000000000000       (mas)\n"
    " pm = proper motion applied? (ok, no)\n"
    " ct = uc (UCAC2); 2m (2MASS); fs (field star); g2 (Gaia-DR2)\n"
    " f = multiplicity flag\n"
    " 0 - no multiple entries per star in astrometry\n"
    " 1 - single position from 2 or more uc/2m entries\n"
    " 2 - single position from 1 uc/2m entry only\n"
    " 3 - fs position from entry with more N contributions\n"
    " 4 - fs position from entry with best (x,y) error\n"
    " 5 - fs position from entry with brightest R mag.\n"
    " 6 - fs position from average over all entries\n"
    " (see details in Assafin et al. 2009)\n"
    " E_ra, E_de: error of star position (mas); (9999 = no estimation)\n"
    " pmra, pmde: star proper motion (mas); (9999 = no proper motion)\n"
    " ----------------------------------------------------------------------------------------------------"
    "--------------------------------------------------------------------------------------------\n"
    "  d  m year  h:m:s UT     ra___dec___J2000_candidate     ra_dec_J2000_target_geocen     C/A    P/A    "
    " vel  Delta  G*   J*   H*   K*   long  loc. t.  off_ra   off_de pm ct f E_ra E_de pmra pmde\n"
    " -----------------------------------------------------------------------------------------------------"
    "-------------------------------------------------------------------------------------------"
    )


ow_occ_head = (
    " Planete: {name}: Star GAIA-DR2, {ephem}\n"
    " Notes:\n"
    " C/A: geocentric closest approach, in arcsec\n"
    " P/A: Planet's position angle wrt to star at C/A, in deg\n"
    " vel: velocity in plane of sky, in km/sec, positive= prograde, negative= retrograde\n"
    " G*, K*: quite rough R and K magnitudes from 1.2m ESO Swiss telescope,\n"
    "         **corrected to a standard velocity of 20 km/sec**\n"
    " Delta: Planet's range to Earth, AU\n"
    " long: East longitude of sub-planet point, deg, positive towards East\n"
    " loc. t.= UT + long: **rough** local solar time at sub-planet point, hh:mm\n"
    " ------------------------------------------------------------------------------------------------"
    "---------------------------------------\n"
    " Selection criteria:\n"
    " Maximum corrected magnitudes R* and K* considered: 100.0\n"
    " Maximum geocentrique closest approach considered: {max_ca:.3f} \n"
    " List below has:          {size}  entries\n"
    " Offset applied to ephemeris (mas):     0.0     0.0\n"
    " {radius}                                                Radius (km)\n"
    " {ow_des}          {ow_des}          {name}             Provisional Designation, Number, Name\n"
    " -------------------------------------------------------------------------------------------"
    "---------------------------------------------------------\n"
    " d  m year  h:m:s UT    ra___dec___J2000_candidate    ra___dec___J2000_ephemeris     C/A    "
    "P/A     vel   Delta   G*   K*   long  loc.t.  e_ra  e_dec\n"
    " -------------------------------------------------------------------------------------------"
    "---------------------------------------------------------\n"
    )



def check_kwargs(input_kwargs, allowed_kwargs, raise_error=True):
    """Tests if the input `**kwargs` are allowed.

    Parameters
    ----------
    input_kwargs : `dict`, `list`
        Dictionary or list with the input values.

    allowed_kwargs : `list`
        List with the allowed keys.

    raise_error : `bool`
        Raises error if ``True``. If ``False``, it will raise the error listing
        the not allowed keys.
    """
    not_allowed = [i for i in input_kwargs if i not in allowed_kwargs]
    if raise_error:
        if len(not_allowed) > 0:
            allowed_kwargs.sort()
            raise TypeError("function got an unexpected keyword argument {}\n"
                            "Available kwargs are: {}".format(not_allowed, allowed_kwargs))
    else:
        return not_allowed


def test_attr(attr, typ, name):
    """Tests attribute.

    This function tests if the attribute ``attr`` belongs to the type ``typ``,
    if not it gives an error message informing the name of the variable.
    """
    try:
        return typ(attr)
    except:
        raise TypeError('"{}" keyword must be a {} object'.format(name, typ))



import os

import astropy.constants as const
import astropy.units as u
import numpy as np
from astropy.coordinates import GCRS, ITRS, SkyOffsetFrame, SkyCoord, EarthLocation, Angle, get_sun
from astropy.time import Time

from sora.config import input_tests

__all__ = ['plot_occ_map']


def xy2latlon(x, y, loncen, latcen, time):
    """Calculates the longitude and latitude given projected positions x and y.

    Parameters
    ----------
    x : `int`, `float`
        Projected position in x, in the GCRS, in meters.

    y : `int`, `float`
        Projected position in y, in the GCRS, in meters.

    loncen : `int`, `float`
        Center longitude of projection, in degrees.

    latcen : `int`, `float`
        Center latitude of projection, in degrees.
    time : `astropy.time.Time`
        Time of referred projection.

    Returns
    -------
    lon, lat : `list`
        Longitude and Latitude whose projection at loncen, lat results
        in x, y. (deg).
    """
    r = const.R_earth.to(u.m).value
    site_cen = EarthLocation(loncen*u.deg, latcen*u.deg)
    itrs_cen = site_cen.get_itrs(obstime=time)
    gcrs_cen = itrs_cen.transform_to(GCRS(obstime=time))

    z = np.array(y, ndmin=1)
    y = np.array(x, ndmin=1)
    x2 = r*r-y*y-z*z

    a = np.where(x2 >= 0.0)
    x = np.sqrt(x2[a])

    y = y[a]
    z = z[a]
    lon = np.repeat(1e+31, len(x2))
    lat = np.repeat(1e+31, len(x2))
    center_frame = SkyOffsetFrame(origin=gcrs_cen)
    if len(x) > 0:
        n = 0
        if not time.isscalar and len(time) == len(x2):
            time = time[a]
        while True:
            n += 1
            new_pos = SkyCoord(x*u.m, y*u.m, z*u.m, representation_type='cartesian', frame=center_frame[a])
            n_coord = new_pos.transform_to(GCRS(obstime=time))
            n_itrs = n_coord.transform_to(ITRS(obstime=time))
            n_site = n_itrs.earth_location
            n_site = EarthLocation(n_site.lon, n_site.lat, 0)
            itrs_site = n_site.get_itrs(obstime=time)
            gcrs_site = itrs_site.transform_to(GCRS(obstime=time))
            target1 = gcrs_site.transform_to(center_frame[a])
            if n == 4:
                lon[a] = n_site.lon.deg
                lat[a] = n_site.lat.deg
                break
            x = target1.cartesian.x.to(u.m).value
    return lon, lat


def latlon2xy(lon, lat, loncen, latcen):
    """Calculates the projection of longitude and latitude in the loncen,
    latcen direction.

    Parameters
    ----------
    lon : `int`, `float`
        Longitude to calculate projection.

    lat : `int`, `float`
        Latitude to calculate projection.

    loncen : `int`, `float`
        Center longitude of projection, in degrees.

    latcen : `int`, `float`
        Center latitude of projection, in degrees.

    Returns
    -------
    x, y : `list`
        Projection of lon, lat at loncen, latcen, in the ITRS (meters).
    """
    site_cen = EarthLocation(loncen*u.deg, latcen*u.deg)
    itrs_cen = site_cen.get_itrs()

    lon = np.array(lon, ndmin=1)
    lat = np.array(lat, ndmin=1)
    site = EarthLocation(lon*u.deg, lat*u.deg, height=0*u.m)
    itrs_site = site.get_itrs()

    target = itrs_site.transform_to(SkyOffsetFrame(origin=itrs_cen))
    y = target.cartesian.y.to(u.m).value
    z = target.cartesian.z.to(u.m).value
    k = np.where(target.cartesian.x.to(u.m).value < 0.0)
    y[k] = 1e+31
    z[k] = 1e+31
    return y, z


def plot_occ_map(name, radius, coord, time, ca, pa, vel, dist, mag=0, longi=0, **kwargs):
    """Plots the map of the occultation.

    Parameters
    ----------
    name : `str`
        Name of the object.

    radius : `int`, `float`
        Radius of the object, in km.

    coord : `str`, `astropy.coordinates.SkyCoord`
        Coordinates of the star (``"hh mm ss.sss dd mm ss.sss"`` or
        ``"hh.hhhhhhhh dd.dddddddd"``).

    time : `str`, `astropy.time.Time`
        Instant of Closest Approach (iso or isot format).

    ca : `int`, `float`
        Closest Approach Distance, in arcsec.

    pa : `int`, `float`
        Position Angle at C/A, in degrees.

    vel : `int`, `float`
        Velocity of the event, in km/s.

    dist : `int`, `float`
        Object distance at C/A, in AU.

    mag : `int`, `float`, default=0
        Mag* = Normalized magnitude to vel=20km/s.

    longi : `int`, `float`, default=0
        East longitude of sub-planet point, deg, positive towards East.

    nameimg : `str`
        Change the name of the imaged saved.

    path : `str`
        Path to a directory where to save map.

    resolution : `int`, default=2
        Cartopy feature resolution.\n
        - ``1`` means a resolution of "10m";\n
        - ``2`` a resolution of "50m";\n
        - ``3`` a resolution of "100m".

    states : `bool`
        If True, plots the states borders of the countries. The states
        of some countries will only be shown depending on the resolution.

    zoom : `int`, `float`
        Zooms in or out of the map.

    centermap_geo : `list`, default=None
        Center the map given coordinates in longitude and latitude. It must be
        a list with two numbers.

    centermap_delta : `list`, default=None
        Displace the center of the map given displacement in X and Y, in km.
        It must be a list with two numbers.

    centerproj : `list`
        Rotates the Earth to show occultation with the center projected at a
        given longitude and latitude. It must be a list with two numbers.

    labels : `bool`, default=True
        Plots text above and below the map with the occultation parameters.

    meridians : `int`, default=30
        Plots lines representing the meridians for given interval, in degrees.

    parallels : `int`, default=30
        Plots lines representing the parallels for given interval, in degrees.

    sites : `dict`
        Plots site positions in map. It must be a python dictionary where the
        key is  the `name` of the site, and the value is a list with `longitude`,
        `latitude`, `delta_x`, `delta_y` and `color`. `delta_x` and `delta_y`
        are displacement, in km, from the point position of the site in the map
        and the `name`. `color` is the color of the point.

    site_name : `bool`
        If True, it prints the name of the sites given, else it plots only the points.

    countries : `dict`
        Plots the names of countries. It must be a python dictionary where the
        key is the name of the country and the value is a list with longitude
        and latitude of the lower left part of the text.

    offset : `list`
        Applies an offset to the ephemeris, calculating new CA and instant of
        CA. It is a pair of `delta_RA*cosDEC` and `delta_DEC`.

    mapstyle : `int`, default=1
        Define the color style of the map. ``'1'`` is the default black
        and white scale. ``'2'`` is a colored map.

    error : `int`, `float`
        Ephemeris error in mas. It plots a dashed line representing radius + error.

    ercolor : `str`
        Changes the color of the lines of the error bar.

    ring : `int`, `float`
        Plots a dashed line representing the location of a ring. It is given
        in km, from the center.

    rncolor : `str`
        Changes the color of ring lines.

    atm : `int`, `float`
        Plots a dashed line representing the location of an atmosphere. It is
        given in km, from the center.

    atcolor : `str`
        Changes the color of atm lines.

    chord_delta : `list`
        List with distances from center to plot chords.

    chord_geo : `2d-list`
        List with pairs of coordinates to plot chords.

    chcolor : `str`, default='grey'
        Color of the line of the chords.

    heights : `list`
        It plots a circular dashed line showing the locations where the observer
        would observe the occultation at a given height above the horizons.
        This must be a list.

    hcolor : `str`
        Changes the color of the height lines.

    mapsize : `list`, default= [46.0, 38.0]
        The size of figure, in cm. It must be a list with two values.

    cpoints : `int`, `float`, default=60
        Interval for the small points marking the center of shadow, in seconds.

    ptcolor : `str`
        Change the color of the center points.

    alpha : `float`, default=0.2
        The transparency of the night shade, where 0.0 is full transparency and
        1.0 is full black.

    fmt : `str`, default:'png'
        The format to save the image. It is parsed directly by `matplotlib.pyplot`.

    dpi : `int`, default=100
        Resolution in "dots per inch". It defines the quality of the image.

    lncolor : `str`
        Changes the color of the line that represents the limits of the shadow
        over Earth.

    outcolor :`str`
        Changes the color of the lines that represents the limits of the shadow
        outside Earth.

    scale : `int`, `float`
        Arbitrary scale for the size of the name of the site.

    cscale : `int`, `float`
        Arbitrary scale for the name of the country.

    sscale : `int`, `float`
        Arbitrary scale for the size of point of the site.

    pscale : `int`, `float`
        Arbitrary scale for the size of the points that represent the center of
        the shadow.

    arrow : `bool`
        If True, it plots the arrow with the occultation direction.


    Important
    ---------
    Required parameters to plot an occultation map: 'name', 'radius', 'coord',
    'time', 'ca', 'pa', 'vel', and 'dist'.


    Note
    ----
    The parameters 'mag' and 'longi' are optional and only printed in label.
    All other remaining parameters can be used to further customize the Map
    configuration.

    When producing the map, only one of 'centermap_geo' or 'centermap_delta'
    options can be used at a time.
    """
    import matplotlib.pyplot as plt
    import cartopy.crs as ccrs
    import cartopy.feature as cfeature

    allowed_kwargs = ['alpha', 'arrow', 'atcolor', 'atm', 'centermap_delta', 'centermap_geo', 'centerproj',
                      'chcolor', 'chord_delta', 'chord_geo', 'countries', 'cpoints', 'cscale', 'dpi', 'ercolor',
                      'error', 'fmt', 'hcolor', 'heights', 'labels', 'lncolor', 'mapsize', 'mapstyle', 'meridians',
                      'nameimg', 'nscale', 'offset', 'outcolor', 'parallels', 'path', 'pscale', 'ptcolor',
                      'resolution', 'ring', 'rncolor', 'site_name', 'sites', 'sscale', 'states', 'zoom']
    input_tests.check_kwargs(kwargs, allowed_kwargs=allowed_kwargs)

    if not type(name) == str:
        raise TypeError('name keyword must be a string')

    radius = radius*u.km

    occs = {}
    try:
        occs['stars'] = SkyCoord(coord, frame='icrs', unit=(u.hourangle, u.degree))
    except:
        raise KeyError('"star" keyword is not in the format: "hh mm ss.sss dd mm ss.sss" or "hh.hhhhhhhh dd.dddddddd"')

    try:
        occs['datas'] = Time(time)
    except:
        raise KeyError('"time" keyword is not a iso or isot time format')
    occs['ca'] = ca*u.arcsec
    occs['posa'] = pa*u.deg
    occs['vel'] = vel*(u.km/u.s)
    occs['dist'] = dist*u.AU
    occs['magG'] = mag
    occs['longi'] = longi

    mapstyle = kwargs.get('mapstyle', 1)
    if mapstyle not in [1, 2]:
        raise ValueError('mapstyle must be 1 or 2]')

    resolution = kwargs.get('resolution', 2)
    if resolution not in [1, 2, 3]:
        raise TypeError('resolution keyword must be one of these: [1, 2, 3] where 1=10m, 2=50m and 3=100m')
    res = ['10m', '50m', '110m']
    resolution = res[resolution-1]

    nameimg = kwargs.get('nameimg', '{}_{}'.format(name, occs['datas'].isot))
    fmt = kwargs.get('fmt', 'png')
    dpi = kwargs.get('dpi', 100)
    step = kwargs.get('step', 1)
    mapsize = kwargs.get('mapsize', [46.0, 38.0])*u.cm
    erro = kwargs.get('error', None)
    ring = kwargs.get('ring', None)
    atm = kwargs.get('atm', None)
    cpoints = kwargs.get('cpoints', 60)
    states = kwargs.get('states', True)
    labels = kwargs.get('labels', True)
    meridians = kwargs.get('meridians', 30)
    parallels = kwargs.get('parallels', 30)
    nscale = kwargs.get('nscale', 1)
    cscale = kwargs.get('cscale', 1)
    sscale = kwargs.get('sscale', 1)
    pscale = kwargs.get('pscale', 1)
    heights = np.array(kwargs.get('heights'), None)
    alpha = kwargs.get('alpha', 0.2)
    centermap_geo = kwargs.get('centermap_geo', None)
    centermap_delta = kwargs.get('centermap_delta', None)
    if 'centermap_geo' in kwargs and 'centermap_delta' in kwargs:
        raise ValueError('User must give "centermap_geo" OR "centermap_delta"')
    zoom = kwargs.get('zoom', 1)
    if zoom <= 0:
        raise ValueError('zoom can not be equal or smaller than 0.')
    off_ra, off_de = kwargs.get('offset', [0.0, 0.0])*u.mas
    arrow = kwargs.get('arrow', True)
    site_name = kwargs.get('site_name', True)
    path = kwargs.get('path', '.')
    if not os.path.exists(path):
        raise IOError('Path does not exists')
    chord_delta = np.array(kwargs.get('chord_delta', []), ndmin=1)*u.km
    chord_geo = kwargs.get('chord_geo', [])
    if len(chord_geo) > 0:
        try:
            b = np.array(chord_geo, ndmin=2)
            chord_geo = b.reshape(len(b), 2)
        except:
            raise ValueError('chord_geo must a set of pairs with longitude and latitude')
        chord_geo = EarthLocation(*chord_geo.T)

    sites = {}
    if 'sites' in kwargs.keys():
        if type(kwargs['sites']) == str and os.path.isfile(kwargs['sites']):
            data = np.loadtxt(kwargs['sites'], dtype={'names': ('name', 'lon', 'lat', 'offx', 'offy', 'color'),
                                                      'formats': ('S30', 'f8', 'f8', 'f8', 'f8', 'S30')},
                              delimiter=',', ndmin=1)
            for i, s in enumerate(data):
                sites[s['name'].strip().decode()] = [s['lon'], s['lat'], s['offx'], s['offy'], s['color'].strip().decode()]
        elif type(kwargs['sites']) == dict:
            sites = kwargs['sites']
        else:
            raise TypeError('sites keyword must be a file or a dictionary')

    countries = {}
    if 'countries' in kwargs.keys():
        if type(kwargs['countries']) == str and os.path.isfile(kwargs['countries']):
            data = np.loadtxt(kwargs['countries'], dtype={'names': ('name', 'lon', 'lat'), 'formats': ('S30', 'f8', 'f8')},
                              delimiter=',', ndmin=1)
            for i, c in enumerate(data):
                countries[c['name'].strip().decode()] = [c['lon'], c['lat']]
        elif type(kwargs['countries']) == dict:
            countries = kwargs['countries']
        else:
            raise TypeError('country keyword must be a file or a dictionary')

    # calculates offsets
    dca = off_ra*np.sin(occs['posa']) + off_de*np.cos(occs['posa'])
    dt = (-(off_ra * np.cos(occs['posa']) - off_de * np.sin(occs['posa'])).to(u.rad) * occs['dist'].to(u.km) / np.absolute(
        occs['vel'])).value * u.s
    ca1 = occs['ca'] + dca
    data = occs['datas'] + dt

    # define map parameters
    center_gcrs = GCRS(occs['stars'].ra, occs['stars'].dec, 1*u.R_earth, obstime=data)
    center_itrs = center_gcrs.transform_to(ITRS(obstime=data))
    center_map = center_itrs.earth_location
    centert = True
    if 'centerproj' in kwargs.keys():
        if type(kwargs['centerproj']) == EarthLocation:
            center_map = kwargs['centerproj']
        elif np.array(kwargs['centerproj']).shape == (2,):
            center_map = EarthLocation.from_geodetic(*kwargs['centerproj'], 0.0)
        else:
            raise TypeError('centerproj must be an Astropy EarthLocation Object or an array with Longitude and Latitude only')
        centert = False
    fig = plt.figure(figsize=(mapsize.to(u.imperial.inch).value))
    projection = ccrs.Orthographic(central_longitude=center_map.lon.value, central_latitude=center_map.lat.value)
    if labels:
        axf = plt.axes(projection=projection)
    else:
        axf = plt.axes([-0.001, -0.001, 1.002, 1.002], projection=projection)
    axf.set_global()

    # calculates regions for zoom
    limits = None
    r = const.R_earth.to(u.m).value
    if centermap_geo is not None:
        cx, cy = latlon2xy(centermap_geo[0], centermap_geo[1], center_map.lon.value, center_map.lat.value)
        limits = [cx/1000.0, cy/1000.0]
        if np.any(np.absolute(limits) > r):
            raise ValueError('Coordinates for centermap_geo are outside the visible range.')
    elif centermap_delta is not None:
        limits = centermap_delta
    elif zoom != 1:
        limits = [0, 0]
    if limits is not None:
        dr = r/zoom
        l0 = (limits[0]*u.km).to(u.m).value
        l1 = (limits[1]*u.km).to(u.m).value
        dmsize = mapsize[0]/mapsize[1]
        if mapsize[1] < mapsize[0]:
            lx = l0 - dr*dmsize
            ux = l0 + dr*dmsize
            ly = l1 - dr
            uy = l1 + dr
        else:
            lx = l0 - dr
            ux = l0 + dr
            ly = l1 - dr/dmsize
            uy = l1 + dr/dmsize
        axf.set_xlim(lx, ux)
        axf.set_ylim(ly, uy)
        if labels and zoom > 1:
            centert = False

    # plots features
    axf.coastlines(resolution=resolution, color='0.3')
    ocean = cfeature.NaturalEarthFeature('physical', 'ocean', resolution)
    land = cfeature.NaturalEarthFeature('physical', 'land', resolution)
    border = cfeature.NaturalEarthFeature('cultural', 'admin_0_countries', resolution)
    if mapstyle == 1:
        axf.add_feature(ocean, zorder=0, color='0.9')
        axf.add_feature(land, zorder=0, edgecolor='None', color='1.0')
        axf.add_feature(border, zorder=0.1, edgecolor='0.4', facecolor='None')
        axf.add_feature(cfeature.RIVERS, zorder=0, edgecolor='0.7')
        axf.add_feature(cfeature.LAKES, zorder=0, color='0.7')
        ptcolor = 'black'
        lncolor = 'blue'
        ercolor = 'blue'
        rncolor = 'blue'
        atcolor = 'blue'
        outcolor = 'red'
        hcolor = 'black'
        chcolor = 'gray'
    elif mapstyle == 2:
        axf.add_feature(ocean, zorder=0, facecolor=cfeature.COLORS['water'])
        axf.add_feature(land, zorder=0, edgecolor='None', facecolor=cfeature.COLORS['land'])
        axf.add_feature(border, zorder=0, edgecolor='0.5', facecolor=cfeature.COLORS['land'])
        axf.add_feature(border, zorder=0.1, edgecolor='0.5', facecolor='None')
        axf.add_feature(cfeature.RIVERS, zorder=0)
        axf.add_feature(cfeature.LAKES, zorder=0)
        ptcolor = 'red'
        lncolor = 'blue'
        ercolor = 'red'
        rncolor = 'black'
        atcolor = 'black'
        outcolor = 'red'
        hcolor = 'black'
        chcolor = 'gray'
    if states:
        states_r = cfeature.NaturalEarthFeature('cultural', 'admin_1_states_provinces', resolution)
        axf.add_feature(states_r, zorder=0, edgecolor='0.6', facecolor='None')

    gl = axf.gridlines(xlocs=np.arange(-180, 180.001, meridians), ylocs=np.arange(-90, 90.001, parallels))
    gl.n_steps = 180
    sun = get_sun(data)
    sun_lat = sun.dec
    sun_lon = sun.ra - data.sidereal_time('mean', 'greenwich')
    pole_lon = sun_lon.deg
    pole_lat = sun_lat.deg
    proj_sun = ccrs.Orthographic(central_longitude=pole_lon+180, central_latitude=-pole_lat)
    bordx = r*np.cos(np.arange(0, 361, 0.5)*u.deg)
    bordy = r*np.sin(np.arange(0, 361, 0.5)*u.deg)
    axf.fill(bordx, bordy, transform=proj_sun, linewidth=0, color='black', alpha=alpha)
    axf.fill(bordx*np.cos(18*u.deg), bordy*np.cos(18*u.deg), transform=proj_sun, linewidth=0, color='black', alpha=alpha)

    ptcolor = kwargs.get('ptcolor', ptcolor)
    lncolor = kwargs.get('lncolor', lncolor)
    ercolor = kwargs.get('ercolor', ercolor)
    rncolor = kwargs.get('rncolor', rncolor)
    atcolor = kwargs.get('atcolor', atcolor)
    outcolor = kwargs.get('outcolor', outcolor)
    hcolor = kwargs.get('hcolor', hcolor)
    chcolor = kwargs.get('chcolor', chcolor)

    # calculates path
    vec = np.arange(0, int(8000/(np.absolute(occs['vel'].value))), step)
    vec = np.sort(np.concatenate((vec, -vec[1:]), axis=0))
    pa = Angle(occs['posa'])
    pa.wrap_at('180d', inplace=True)
    if pa > 90*u.deg:
        paplus = pa - 180*u.deg
    elif pa < -90*u.deg:
        paplus = pa + 180*u.deg
    else:
        paplus = pa
    deltatime = vec*u.s
    datas1 = data + deltatime
    centers_gcrs = GCRS(np.repeat(occs['stars'].ra, len(datas1)), np.repeat(occs['stars'].dec, len(datas1)),
                        1*u.R_earth, obstime=datas1)
    centers_itrs = centers_gcrs.transform_to(ITRS(obstime=datas1))
    centers = centers_itrs.earth_location

    dista = (occs['dist'].to(u.km)*ca1.to(u.rad)).value*u.km
    ax = dista*np.sin(pa) + (deltatime*occs['vel'])*np.cos(paplus)
    by = dista*np.cos(pa) - (deltatime*occs['vel'])*np.sin(paplus)

    ax2 = ax - radius * np.sin(paplus)
    by2 = by - radius * np.cos(paplus)
    lon1, lat1 = xy2latlon(ax2.to(u.m).value, by2.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
    j = np.where(lon1 < 1e+30)
    axf.plot(lon1[j], lat1[j], transform=ccrs.Geodetic(), color=lncolor)
    j = np.where(lon1 > 1e+30)
    if 'centerproj' not in kwargs:
        plt.plot(ax2[j].to(u.m).value, by2[j].to(u.m).value, color=outcolor, clip_on=(not centert), zorder=-0.2)

    ax3 = ax + radius * np.sin(paplus)
    by3 = by + radius * np.cos(paplus)
    lon2, lat2 = xy2latlon(ax3.to(u.m).value, by3.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
    j = np.where(lon2 < 1e+30)
    axf.plot(lon2[j], lat2[j], transform=ccrs.Geodetic(), color=lncolor)
    j = np.where(lon2 > 1e+30)
    if 'centerproj' not in kwargs:
        plt.plot(ax3[j].to(u.m).value, by3[j].to(u.m).value, color=outcolor, clip_on=(not centert), zorder=-0.2)

    # plots chords_delta
    for val in chord_delta:
        ax2 = ax + val*np.sin(paplus)
        by2 = by + val*np.cos(paplus)
        lon1, lat1 = xy2latlon(ax2.to(u.m).value, by2.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon1 < 1e+30)
        axf.plot(lon1[j], lat1[j], transform=ccrs.Geodetic(), color=chcolor)

    # plots chords_geo
    for coord_geo in chord_geo:
        xt, yt = latlon2xy(coord_geo.lon.deg, coord_geo.lat.deg, centers.lon.value, centers.lat.value)*u.m
        val = np.sqrt((xt-ax)**2 + (yt-by)**2)
        k = val.argmin()
        ang = np.arctan2((yt-by)[k], (xt-ax)[k])
        val = np.sign(np.sin(ang))*val[k]
        ax2 = ax + val*np.sin(paplus)
        by2 = by + val*np.cos(paplus)
        lon1, lat1 = xy2latlon(ax2.to(u.m).value, by2.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon1 < 1e+30)
        axf.plot(lon1[j], lat1[j], transform=ccrs.Geodetic(), color=chcolor)

    # plots error
    if erro is not None:
        err = erro*u.mas
        errd = (occs['dist'].to(u.km)*err.to(u.rad)).value*u.km
        ax2 = ax - errd*np.sin(paplus) - radius*np.sin(paplus)
        by2 = by - errd*np.cos(paplus) - radius*np.cos(paplus)
        lon1, lat1 = xy2latlon(ax2.to(u.m).value, by2.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon1 < 1e+30)
        axf.plot(lon1[j], lat1[j], '--', transform=ccrs.Geodetic(), color=ercolor)

        ax3 = ax + errd*np.sin(paplus) + radius*np.sin(paplus)
        by3 = by + errd*np.cos(paplus) + radius*np.cos(paplus)
        lon2, lat2 = xy2latlon(ax3.to(u.m).value, by3.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon2 < 1e+30)
        axf.plot(lon2[j], lat2[j], '--', transform=ccrs.Geodetic(), color=ercolor)

    # plots ring
    if ring is not None:
        rng = ring*u.km
        ax2 = ax - rng*np.sin(paplus)
        by2 = by - rng*np.cos(paplus)
        lon1, lat1 = xy2latlon(ax2.to(u.m).value, by2.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon1 < 1e+30)
        axf.plot(lon1[j], lat1[j], '--', transform=ccrs.Geodetic(), color=rncolor)

        ax3 = ax + rng*np.sin(paplus)
        by3 = by + rng*np.cos(paplus)
        lon2, lat2 = xy2latlon(ax3.to(u.m).value, by3.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon2 < 1e+30)
        axf.plot(lon2[j], lat2[j], '--', transform=ccrs.Geodetic(), color=rncolor)

    # plots atm
    if atm is not None:
        atmo = atm*u.km
        ax2 = ax - atmo*np.sin(paplus)
        by2 = by - atmo*np.cos(paplus)
        lon1, lat1 = xy2latlon(ax2.to(u.m).value, by2.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon1 < 1e+30)
        axf.plot(lon1[j], lat1[j], '--', transform=ccrs.Geodetic(), color=atcolor)

        ax3 = ax + atmo*np.sin(paplus)
        by3 = by + atmo*np.cos(paplus)
        lon2, lat2 = xy2latlon(ax3.to(u.m).value, by3.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon2 < 1e+30)
        axf.plot(lon2[j], lat2[j], '--', transform=ccrs.Geodetic(), color=atcolor)

# plots center points
    vec = np.arange(0, int(8000/(np.absolute(occs['vel'].value))), cpoints)
    deltatime = np.sort(np.concatenate((vec, -vec[1:]), axis=0))*u.s
    axc = dista*np.sin(pa) + (deltatime*occs['vel'])*np.cos(paplus)
    byc = dista*np.cos(pa) - (deltatime*occs['vel'])*np.sin(paplus)
    plt.plot(axc.to(u.m).value, byc.to(u.m).value, 'o', color=ptcolor, clip_on=(not centert),
             markersize=mapsize[0].value*pscale*8.0/46.0, zorder=-0.2)

    datas2 = data + deltatime
    centers_p_gcrs = GCRS(np.repeat(occs['stars'].ra, len(datas2)), np.repeat(occs['stars'].dec, len(datas2)),
                          1*u.R_earth, obstime=datas2)
    centers_p_itrs = centers_p_gcrs.transform_to(ITRS(obstime=datas2))
    centers_p = centers_p_itrs.earth_location
    clon1, clat1 = xy2latlon(axc.to(u.m).value, byc.to(u.m).value, centers_p.lon.value, centers_p.lat.value, datas2)
    j = np.where(clon1 < 1e+30)
    axf.plot(clon1[j], clat1[j], 'o', transform=ccrs.Geodetic(), color=ptcolor, clip_on=True,
             markersize=mapsize[0].value*pscale*8.0/46.0)

    datas1 = data + deltatime
    center_gcrs = GCRS(np.repeat(occs['stars'].ra, 1), np.repeat(occs['stars'].dec, 1),
                       1*u.R_earth, obstime=data)
    center_itrs = center_gcrs.transform_to(ITRS(obstime=data))
    center = center_itrs.earth_location
    xp = [(dista.to(u.m)*np.sin(pa)).value]
    yp = [(dista.to(u.m)*np.cos(pa)).value]
    loncen, latcen = xy2latlon(xp, yp, center.lon.value, center.lat.value, data)
    j = np.where(loncen < 1e+30)
    if len(j) > 0:
        axf.plot(loncen[j], latcen[j], 'o', transform=ccrs.Geodetic(), color=ptcolor, clip_on=True,
                 markersize=mapsize[0].value*pscale*24.0/46.0)
    elif not centert:
        plt.plot(xp, yp, 'o', color=ptcolor, clip_on=False, markersize=mapsize[0].value*pscale*24.0/46.0)

    # plots the heights
    if 'heights' in kwargs.keys():
        for h in heights:
            lonb, latb = xy2latlon(bordx * np.cos(h * u.deg), bordy * np.cos(h * u.deg), center.lon.value,
                                   center.lat.value, data)
            axf.plot(lonb, latb, transform=ccrs.Geodetic(), linestyle='dotted', color=hcolor)

    # plots the the direction arrow
    if arrow:
        if limits is None:
            plt.quiver(5500000, -5500000, (np.sin(paplus+90*u.deg)*np.sign(occs['vel'])).value,
                       (np.cos(paplus+90*u.deg)*np.sign(occs['vel'])).value, width=0.005)
        else:
            plt.quiver(lx + (ux-lx)*0.9, ly + (uy-ly)*0.1, (np.sin(paplus+90*u.deg)*np.sign(occs['vel'])).value,
                       (np.cos(paplus+90*u.deg)*np.sign(occs['vel'])).value, width=0.005, zorder=1.3)

    # plots the countries names
    for country in countries.keys():
        plt.text(countries[country][0], countries[country][1], country, transform=ccrs.Geodetic(),
                 weight='bold', color='grey', fontsize=30*cscale, family='monospace')

    # plots the sites
    for site in sites.keys():
        s = EarthLocation.from_geodetic(sites[site][0], sites[site][1], 0.0*u.km)
        axf.plot(s.lon.deg, s.lat.deg, 'o', transform=ccrs.Geodetic(),
                 markersize=mapsize[0].value*sscale*10.0/46.0, color=sites[site][4])
        if site_name:
            xt, yt = latlon2xy(s.lon.deg, s.lat.deg, center_map.lon.value, center_map.lat.value)
            axf.text(xt + sites[site][2]*1000, yt+sites[site][3]*1000, site, weight='bold',
                     fontsize=25*nscale, family='monospace')

    # Define the title and label of the output
    title = ('Object        Diam   Tmax   dots <> ra_offset_dec\n'
             '{:10s} {:4.0f} km  {:5.1f}s  {:02d} s <>{:+6.1f} {:+6.1f} \n'.
             format(name, 2*radius.value, (2*radius/np.absolute(occs['vel'])).value,
                    cpoints, off_ra.value, off_de.value))
    labelx = ("\n year-m-d    h:m:s UT     ra__dec__J2000__candidate    C/A    P/A    vel   Delta   G*  long\n"
              "{}  {:02d} {:02d} {:07.4f} {:+03d} {:02d} {:06.3f} {:6.3f} {:6.2f} {:6.2f}  {:5.2f} {:5.1f}  {:3.0f}".
              format(data.iso, int(occs['stars'].ra.hms.h), int(occs['stars'].ra.hms.m), occs['stars'].ra.hms.s,
                     int(occs['stars'].dec.dms.d), np.absolute(int(occs['stars'].dec.dms.m)),
                     np.absolute(occs['stars'].dec.dms.s), ca1.value, occs['posa'].value,
                     occs['vel'].value, occs['dist'].value, occs['magG'], occs['longi']))

    # plots the map
    if labels:
        axf.set_title(title, family='monospace', weight='bold', fontsize=22)
        axf.text(0.5, -0.1, labelx, va='bottom', ha='center', rotation='horizontal', rotation_mode='anchor',
                 transform=axf.transAxes, family='monospace', weight='bold', fontsize=22)
    filepath = os.path.join(path, '{}.{}'.format(nameimg, fmt))
    plt.savefig(filepath, format=fmt, dpi=dpi)
    print('{}.{} generated'.format(nameimg, fmt))
    plt.clf()
    plt.close()



import warnings

import astropy.units as u
import numpy as np
from astropy.time import Time

from sora.config.decorators import deprecated_function, deprecated_alias
from sora.prediction import occ_params, PredictionTable
from .fitting import fit_ellipse as ellipse_fitting

__all__ = ['Occultation']
warnings.simplefilter('always', UserWarning)


class Occultation:
    """Instantiates the Occultation object and performs the reduction of the
    occultation.

    Attributes
    ----------
    star : `sora.Star`, `str`, required
        the coordinate of the star in the same reference frame as the ephemeris.
        It must be a Star object or a string with the coordinates of the object
        to search on Vizier.

    body : `sora.Body`, `str`
        Object that will occult the star. It must be a Body object or its name
        to search in the Small Body Database.

    ephem : `sora.Ephem`, `list`
        Object ephemeris. It must be an Ephemeris object or a list.

    time : `str`, `astropy.time.Time`, required
        Reference time of the occultation. Time does not need to be exact, but
        needs to be within approximately 50 minutes of the occultation closest
        approach to calculate occultation parameters.

    reference_center : `str`, `sora.Observer`, `sora.Spacecraft`
        A SORA observer object or a string 'geocenter'.
        The occultation parameters will be calculated in respect
        to this reference as center of projection.


    Important
    ---------
    When instantiating with "body" and "ephem", the user may define the
    Occultation in 3 ways:

    1. With `body` and `ephem`.

    2. With only "body". In this case, the "body" parameter must be a Body
    object and have an ephemeris associated (see Body documentation).

    3. With only `ephem`. In this case, the `ephem` parameter must be one of the
    Ephem Classes and have a name (see Ephem documentation) to search for the
    body in the Small Body Database.

    """
    def __init__(self, star, body=None, ephem=None, time=None, reference_center='geocenter'):

        from sora.body import Body
        from sora.star import Star
        from .chordlist import ChordList

        if body is None and ephem is None:
            raise ValueError('"body" and/or "ephem" must be given.')
        if time is None:
            raise ValueError('"time" parameter must be given.')
        if isinstance(star, str):
            star = Star(coord=star)
        elif not isinstance(star, Star):
            raise ValueError('"star" must be a Star object or a string with coordinates of the star')
        self._star = star
        if body is not None:
            if not isinstance(body, (str, Body)):
                raise ValueError('"body" must be a string with the name of the object or a Body object')
            if isinstance(body, str):
                body = Body(name=body)
            self._body = body
        if ephem is not None:
            if body is not None:
                self.body.ephem = ephem
            else:
                if hasattr(ephem, 'name'):
                    self._body = Body(name=ephem.name, ephem=ephem)
                else:
                    raise ValueError('When only "ephem" is given, "ephem" must have a name for search.')
        try:
            ephem = self.body.ephem
        except AttributeError:
            raise ValueError('An Ephem object must be defined in Body object.')
        self._reference_center = reference_center

        tca, ca, pa, vel, dist = occ_params(self.star, ephem, time, reference_center=reference_center)
        self.ca = ca   # Closest Approach distance
        self.pa = pa   # Position Angle at CA
        self.vel = vel  # Shadow velocity at CA
        self.dist = dist  # object distance at CA
        self.tca = tca   # Instant of CA
        self.star_diam = self.star.apparent_diameter(self.dist, verbose=False)

        meta = {
            'name': self.body.name, 'radius': self.body.radius.to(u.km).value,
            'error_ra': self.body.ephem.error_ra.to(u.mas).value, 'error_dec': self.body.ephem.error_dec.to(u.mas).value}
        self.predict = PredictionTable(
            time=[tca], coord_star=[self.star.get_position(tca, observer=reference_center)],
            coord_obj=[self.body.ephem.get_position(tca, observer=reference_center)],
            ca=[ca.value], pa=[pa.value], vel=[vel.value], dist=[dist.value],
            mag=[self.star.mag['G']], source=[self.star.code], meta=meta)

        self.__observations = []
        self._chords = ChordList(star=self.star, body=self._body, time=self.tca)
        self._chords._shared_with['occultation'] = {"vel": np.absolute(self.vel), "dist": float(self.dist.AU),
                                                    "star_diam": float(self.star_diam.km)}

    @property
    def star(self):
        return self._star

    @property
    def body(self):
        return self._body

    @property
    def chords(self):
        return self._chords

    # remove this block for v1.0
    @deprecated_function(message="Please use chords.add_chord to add new observations.")
    def add_observation(self, obs, lightcurve):
        """Adds observations to the Occultation object.

        Parameters
        ----------
        obs : `sora.Observer`
            The Observer object to be added.

        lightcurve : `sora.LightCurve`
            The LightCurve object to be added.
        """
        self.chords.add_chord(name=lightcurve.name, observer=obs, lightcurve=lightcurve)

    @deprecated_function(message="Please use chords.remove_chord to remove observations.")
    def remove_observation(self, key, key_lc=None):
        """Removes an observation from the Occultation object.

        Parameters
        ----------
        key : `str`
            The name given to `Observer` or `LightCurve` to remove from the list.

        key_lc : `str`
            In the case where repeated names are present for different
            observations, `key_lc` must be given for the name of the `LightCurve`
            and key will be used for the name of the `Observer`.
        """
        try:
            self.chords.remove_chord(name=key)
        except KeyError:
            rm_list = []
            for name, chord in self.chords.items():
                if chord.observer.name == key and (key_lc is None or chord.lightcurve.name == key_lc):
                    rm_list.append(name)
            if len(rm_list) != 1:
                if len(rm_list) == 0:
                    err_mes = "No observation was identified with given keys."
                else:
                    err_mes = "More than one chord was identified. Please provide unique keys."
                raise ValueError(err_mes)
            else:
                self.chords.remove_chord(name=rm_list[0])

    @deprecated_function(message="Please use chords.")
    def observations(self):
        """ Print all the observations added to the Occultation object
        Pair (`Observer`, `LightCurve`)
        """
        print(self.chords.__repr__())

    # end of block removal

    def fit_ellipse(self, **kwargs):
        chisquare = ellipse_fitting(self, **kwargs)
        return chisquare

    fit_ellipse.__doc__ = ellipse_fitting.__doc__

    # remove this block for v1.0
    @property
    @deprecated_function(message="Please use chords.summary()")
    def positions(self):
        """ Calculates the position and velocity for all chords.
        Saves it into an `_PositionDict` object.
        """
        from functools import partial
        from .meta import _PositionDict

        if not hasattr(self, '_position'):
            self._position = _PositionDict()
        position = self._position
        if len(self.chords) == 0:
            raise ValueError('There is no observation defined for this occultation')

        pair = []
        for name, chord in self.chords.items():
            obs = chord.observer
            obs_name = obs.name
            if obs_name == '':
                obs_name = '{}_obs'.format(chord.name)
            lc = chord.lightcurve
            lc_name = lc.name
            if lc_name == '':
                lc_name = '{}_lc'.format(chord.name)
            pair.append((obs_name, lc_name))

            coord = [obs.lon, obs.lat, obs.height]
            if obs.name not in position.keys():
                position['_occ_'+obs_name] = _PositionDict(lon=obs.lon, lat=obs.lat, height=obs.height)
                position[obs_name]['_occ_lon'] = obs.lon
                position[obs_name]['_occ_lat'] = obs.lat
                position[obs_name]['_occ_height'] = obs.height
            pos_obs = position[obs_name]
            coord2 = [pos_obs['lon'], pos_obs['lat'], pos_obs['height']]
            if obs.lon != pos_obs['lon']:
                position[obs_name]['_occ_lon'] = obs.lon
            if obs.lat != pos_obs['lat']:
                position[obs_name]['_occ_lat'] = obs.lat
            if obs.height != pos_obs['height']:
                position[obs_name]['_occ_height'] = obs.height
            samecoord = (coord == coord2)

            if lc_name not in pos_obs.keys():
                pos_obs['_occ_'+lc_name] = _PositionDict()
            pos_lc = pos_obs[lc_name]

            pos_lc['_occ_status'] = chord.status()

            if hasattr(lc, 'immersion'):
                if 'immersion' not in pos_lc.keys():
                    pos_lc['_occ_immersion'] = _PositionDict(on=chord.is_able['immersion'],
                                                             enable=partial(chord.enable, time='immersion'),
                                                             disable=partial(chord.disable, time='immersion'))
                obs_im = pos_lc['immersion']
                obs_im['_occ_on'] = chord.is_able['immersion']
                do_err = False
                if samecoord and 'time' in obs_im.keys() and obs_im['time'] == lc.immersion:
                    pass
                else:
                    do_err = True
                    f1, g1, vf1, vg1 = chord.get_fg(time='immersion', vel=True)
                    obs_im['_occ_time'] = lc.immersion
                    obs_im['_occ_value'] = (round(f1, 3), round(g1, 3))
                    obs_im['_occ_vel'] = (round(vf1, 3), round(vg1, 3))
                if not do_err and 'time_err' in obs_im.keys() and obs_im['time_err'] == lc.immersion_err:
                    pass
                else:
                    fe1, ge1 = chord.get_fg(time=lc.immersion-lc.immersion_err*u.s)
                    fe2, ge2 = chord.get_fg(time=lc.immersion+lc.immersion_err*u.s)
                    obs_im['_occ_time_err'] = lc.immersion_err
                    obs_im['_occ_error'] = ((round(fe1, 3), round(ge1, 3)), (round(fe2, 3), round(ge2, 3)))

            if hasattr(lc, 'emersion'):
                pos_lc['_occ_emersion'] = _PositionDict(on=chord.is_able['emersion'],
                                                        enable=partial(chord.enable, time='emersion'),
                                                        disable=partial(chord.disable, time='emersion'))
                obs_em = pos_lc['emersion']
                do_err = False
                if samecoord and 'time' in obs_em.keys() and obs_em['time'] == lc.emersion:
                    pass
                else:
                    do_err = True
                    f1, g1, vf1, vg1 = chord.get_fg(time='emersion', vel=True)
                    obs_em['_occ_time'] = lc.emersion
                    obs_em['_occ_value'] = (round(f1, 3), round(g1, 3))
                    obs_em['_occ_vel'] = (round(vf1, 3), round(vg1, 3))
                if not do_err and 'time_err' in obs_em.keys() and obs_em['time_err'] == lc.emersion_err:
                    pass
                else:
                    fe1, ge1 = chord.get_fg(time=lc.emersion-lc.emersion_err*u.s)
                    fe2, ge2 = chord.get_fg(time=lc.emersion+lc.emersion_err*u.s)
                    obs_em['_occ_time_err'] = lc.emersion_err
                    obs_em['_occ_error'] = ((round(fe1, 3), round(ge1, 3)), (round(fe2, 3), round(ge2, 3)))

            if pos_lc['status'] == 'negative':
                if 'start_obs' not in pos_lc.keys():
                    pos_lc['_occ_start_obs'] = _PositionDict()
                obs_start = pos_lc['start_obs']
                if samecoord and 'time' in obs_start.keys() and obs_start['time'] == lc.initial_time:
                    pass
                else:
                    f, g, vf, vg = chord.get_fg(time='start', vel=True)
                    obs_start['_occ_time'] = lc.initial_time
                    obs_start['_occ_value'] = (round(f, 3), round(g, 3))
                    obs_start['_occ_vel'] = (round(vf, 3), round(vg, 3))
                if 'end_obs' not in pos_lc.keys():
                    pos_lc['_occ_end_obs'] = _PositionDict()
                obs_end = pos_lc['end_obs']
                if samecoord and 'time' in obs_end.keys() and obs_end['time'] == lc.end_time:
                    pass
                else:
                    f, g, vf, vg = chord.get_fg(time='end', vel=True)
                    obs_end['_occ_time'] = lc.end_time
                    obs_end['_occ_value'] = (round(f, 3), round(g, 3))
                    obs_end['_occ_vel'] = (round(vf, 3), round(vg, 3))

        for key in list(position):
            n = 0
            for key_lc in list(position[key]):
                if type(position[key][key_lc]) != _PositionDict:
                    continue
                if (key, key_lc) not in pair:
                    del position[key][key_lc]
                else:
                    n += 1
            if n == 0:
                del position[key]

        return self._position

    @positions.setter
    def positions(self, value):
        """
        Note
        ----
        If the users tries to set a value to position, it must be ``'on'`` or
        ``'off'``, and it will be assigned to all chords.
        """
        if value not in ['on', 'off']:
            raise ValueError("Value must be 'on' or 'off' only.")
        pos = self.positions
        for key in pos.keys():
            pos[key] = value

    # end of block removal

    def check_velocities(self):
        """Prints the current velocity used by the LightCurves and its radial velocity.
        """
        if hasattr(self, 'fitted_params'):
            center = np.array([self.fitted_params['center_f'][0], self.fitted_params['center_g'][0]])
        else:
            center = np.array([0, 0])
        for name, chord in self.chords.items():
            im = getattr(chord.lightcurve, 'immersion', None)
            em = getattr(chord.lightcurve, 'emersion', None)
            if im is None and em is None:
                continue
            print('{} - Velocity used: {:.3f}'.format(name, chord.lightcurve.vel))
            if im is not None:
                vals = chord.get_fg(time=im, vel=True)
                delta = np.array(vals[0:2]) - center
                print('    Immersion Radial Velocity: {:.3f}'.
                      format(np.abs(np.dot(np.array(vals[2:]), delta)/np.linalg.norm(delta))))
            if em is not None:
                vals = chord.get_fg(time=em, vel=True)
                delta = np.array(vals[0:2]) - center
                print('    Emersion Radial Velocity: {:.3f}'.
                      format(np.abs(np.dot(np.array(vals[2:]), delta)/np.linalg.norm(delta))))

    @deprecated_alias(log='verbose')  # remove this line in v1.0
    def new_astrometric_position(self, time=None, offset=None, error=None, verbose=True, observer=None):
        """Calculates the new astrometric position for the object given fitted parameters.

        Parameters
        ----------
        time : `str`, `astropy.time.Time`
            Reference time to calculate the position. If not given, it uses the
            instant of the occultation Closest Approach. It can be a string
            in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.

        offset : `list`
            Offset to apply to the position. If not given, uses the parameters
            from the fitted ellipse.

            Note
            ----
            Must be a list of 3 values being [X, Y, 'unit']. 'unit' must be
            Angular or Distance unit.

            If Distance units for X and Y:
            Ex: [100, -200, 'km'], [0.001, 0.002, 'AU']

            If Angular units fox X [d*a*cos(dec)] and Y [d*dec]:
            Ex: [30.6, 20, 'mas'], or [-15, 2, 'arcsec']

        error : `list`
            Error bar of the given offset. If not given, it uses the 1-sigma
            value of the fitted ellipse.

            Note
            ----
            Error must be a list of 3 values being [dX, dY, 'unit'], similar to
            offset. It does not need to be in the same unit as offset.

        verbose : `bool`
            If true, it Prints text, else it Returns text.

        observer : `str`, `sora.Observer`, `sora.Spacecraft`
            IAU code of the observer (must be present in given list of kernels),
            a SORA observer object or a string: ['geocenter', 'barycenter']
        """
        from astropy.coordinates import SkyCoord, SkyOffsetFrame

        if time is not None:
            time = Time(time)
        else:
            time = self.tca

        tca_diff = np.absolute(time-self.tca)
        if tca_diff > 1*u.day:
            warnings.warn('The difference between the given time and the closest approach instant is {:.1f} days. '
                          'This position could not have a physical meaning.'.format(tca_diff.jd))

        if offset is not None:
            off_ra = offset[0]*u.Unit(offset[2])
            off_dec = offset[1]*u.Unit(offset[2])
            if off_ra.unit.physical_type == 'length':
                dist = True
            elif off_ra.unit.physical_type == 'angle':
                dist = False
            else:
                raise ValueError('Offset unit must be a distance or angular value.')
        elif hasattr(self, 'fitted_params'):
            off_ra = self.fitted_params['center_f'][0]*u.km
            off_dec = self.fitted_params['center_g'][0]*u.km
            dist = True
        else:
            warnings.warn('No offset given or found. Using 0.0 instead.')
            off_ra = 0.0*u.mas
            off_dec = 0.0*u.mas
            dist = False

        if error is not None:
            e_off_ra = error[0]*u.Unit(error[2])
            e_off_dec = error[1]*u.Unit(error[2])
            if e_off_ra.unit.physical_type == 'length':
                e_dist = True
            elif e_off_ra.unit.physical_type == 'angle':
                e_dist = False
            else:
                raise ValueError('Error unit must be a distance or angular value.')
        elif hasattr(self, 'fitted_params'):
            e_off_ra = self.fitted_params['center_f'][1]*u.km
            e_off_dec = self.fitted_params['center_g'][1]*u.km
            e_dist = True
        else:
            warnings.warn('No error given or found. Using 0.0 instead.')
            e_off_ra = 0.0*u.mas
            e_off_dec = 0.0*u.mas
            e_dist = False

        if observer is None:
            observer = self._reference_center

        coord = self.body.ephem.get_position(time, observer=observer)
        distance = coord.distance.to(u.km)
        coord_frame = SkyOffsetFrame(origin=coord)
        if dist:
            off_ra = np.arctan2(off_ra, distance)
            off_dec = np.arctan2(off_dec, distance)
        if e_dist:
            e_off_ra = np.arctan2(e_off_ra, distance)
            e_off_dec = np.arctan2(e_off_dec, distance)
        new_pos = SkyCoord(lon=off_ra, lat=off_dec, frame=coord_frame)
        new_pos = new_pos.icrs

        error_star = self.star.error_at(self.tca)
        error_ra = np.sqrt(error_star[0]**2 + e_off_ra**2)
        error_dec = np.sqrt(error_star[1]**2 + e_off_dec**2)

        out = 'Ephemeris offset (km): X = {:.1f} +/- {:.1f}; Y = {:.1f} +/- {:.1f}\n'.format(
            distance*np.sin(off_ra.to(u.mas)).value, distance*np.sin(e_off_ra.to(u.mas)).value,
            distance*np.sin(off_dec.to(u.mas)).value, distance*np.sin(e_off_dec.to(u.mas)).value)
        out += 'Ephemeris offset (mas): da_cos_dec = {:.3f} +/- {:.3f}; d_dec = {:.3f} +/- {:.3f}\n'.format(
            off_ra.to(u.mas).value, e_off_ra.to(u.mas).value, off_dec.to(u.mas).value, e_off_dec.to(u.mas).value)
        out += '\nAstrometric object position at time {} for reference {}\n'.format(time.iso, observer.__repr__())
        out += 'RA = {} +/- {:.3f} mas; DEC = {} +/- {:.3f} mas'.format(
            new_pos.ra.to_string(u.hourangle, precision=7, sep=' '), error_ra.to(u.mas).value,
            new_pos.dec.to_string(u.deg, precision=6, sep=' '), error_dec.to(u.mas).value)

        if verbose:
            print(out)
        else:
            return out

    # remove this block for v1.0
    @deprecated_function(message="Please use chords.plot_chords to have a better control of the plots")
    def plot_chords(self, all_chords=True, positive_color='blue', negative_color='green', error_color='red',
                    ax=None, lw=2):
        """Plots the chords of the occultation.

        Parameters
        ----------
        all_chords : `bool`, default=True
            If True, it plots all the chords. If False, it sees what was
            deactivated in self.positions and ignores them.

        positive_color : `str`, default='blue'
            Color for the positive chords.

        negative_color : `str`, default='green'
            Color for the negative chords.

        error_color : `str`, default='red'
            Color for the error bars of the chords.

        ax : `maptlotlib.pyplot.Axes`, default=None
            Axis where to plot chords (default: Uses matplotlib pool).

        lw : `int`, `float`, default=2
            Linewidth of the chords.

        """
        self.chords.plot_chords(segment='positive', only_able=not all_chords, color=positive_color, lw=lw, ax=ax)
        self.chords.plot_chords(segment='error', only_able=not all_chords, color=error_color, lw=lw, ax=ax)
        self.chords.plot_chords(segment='negative', color=negative_color, lw=lw, ax=ax, linestyle='--')

    # end of block removal

    def get_map_sites(self):
        """Returns Dictionary with sites in the format required by plot_occ_map function.

        Returns
        -------
        sites : `dict`
            Dictionary with the sites in the format required by `plot_occ_map`
            function.
        """
        sites = {}
        color = {'positive': 'blue', 'negative': 'red'}
        for name, chord in self.chords.items():
            obs = chord.observer
            sites[name] = [obs.lon.deg, obs.lat.deg, 10, 10, color[chord.status()]]
        return sites

    def plot_occ_map(self, **kwargs):
        if 'radius' not in kwargs and hasattr(self, 'fitted_params'):
            r_equa = self.fitted_params['equatorial_radius'][0]
            obla = self.fitted_params['oblateness'][0]
            pos_ang = self.fitted_params['position_angle'][0]
            theta = np.linspace(-np.pi, np.pi, 1800)
            map_pa = self.predict['P/A'][0]
            circle_x = r_equa*np.cos(theta)
            circle_y = r_equa*(1.0-obla)*np.sin(theta)
            ellipse_y = -circle_x*np.sin((pos_ang-map_pa)*u.deg) + circle_y*np.cos((pos_ang-map_pa)*u.deg)
            kwargs['radius'] = ellipse_y.max()
            print('Projected shadow radius = {:.1f} km'.format(kwargs['radius']))
        kwargs['sites'] = kwargs.get('sites', self.get_map_sites())
        if 'offset' not in kwargs and hasattr(self, 'fitted_params'):
            off_ra = self.fitted_params['center_f'][0]*u.km
            off_dec = self.fitted_params['center_g'][0]*u.km
            off_ra = np.arctan2(off_ra, self.dist)
            off_dec = np.arctan2(off_dec, self.dist)
            kwargs['offset'] = [off_ra, off_dec]
        self.predict.plot_occ_map(**kwargs)

    plot_occ_map.__doc__ = PredictionTable.plot_occ_map.__doc__

    def to_log(self, namefile=None):
        """Saves the occultation log to a file.

        Parameters
        ----------
        namefile : `str`
            Filename to save the log.
        """
        if namefile is None:
            namefile = 'occ_{}_{}.log'.format(self.body.shortname.replace(' ', '_'), self.tca.isot[:16])
        f = open(namefile, 'w')
        f.write(self.__str__())
        f.close()

    def check_time_shift(self, time_interval=30, time_resolution=0.001, verbose=False, plot=False, use_error=True, delta_plot=100,
                         ignore_chords=None):
        """Check the needed time offset, so all chords have their center aligned.

        Parameters
        ----------
        time_interval : `int`, `float`
            Time interval to check, default is 30 seconds.

        time_resolution : `int`, `float`
            Time resolution of the search, default is 0.001 seconds.

        verbose : `bool`
            If True, it prints text, default is False.

        plot : `bool`, default=False
            If True, it plots figures as a visual aid.

        use_error : `bool`, default=True
            if True, the linear fit considers the time uncertainty.

        delta_plot : `int`, `float`, default=100
            Value to be added to increase the plot limit, in km.

        ignore_chords : `str`, list, default=None
            Names of the chords to be ignored in the linear fit.

        Returns
        -------
        time_shift : `dict`
            Dictionary with needed time offset to align the chords, each key is
            the name of the chord.
        """
        import matplotlib.pyplot as plt
        fm = np.array([])
        gm = np.array([])
        dfm = np.array([])
        dgm = np.array([])
        vfm = np.array([])
        vgm = np.array([])
        chord_name = np.array([])
        ignore_chords = np.array(ignore_chords, ndmin=1)
        use_chords = np.array([], dtype=bool)
        out_dic = {}

        chords = range(len(self.chords))
        for i in chords:
            chord = self.chords[i]
            if chord.status() == 'positive':
                ffm, ggm, vffm, vggm = chord.get_fg(time=chord.lightcurve.time_mean, vel=True)
                df1, dg1, df2, dg2 = chord.path(segment='error')
                dfm = np.append(dfm, np.sqrt(((df1.max() - df1.min())/2)**2 + ((df2.max() - df2.min())/2)**2))
                dgm = np.append(dgm, np.sqrt(((dg1.max() - dg1.min())/2)**2 + ((dg2.max() - dg2.min())/2)**2))
                fm = np.append(fm, ffm)
                gm = np.append(gm, ggm)
                vfm = np.append(vfm, vffm)
                vgm = np.append(vgm, vggm)
                chord_name = np.append(chord_name, chord.name)
                if chord.name in ignore_chords:
                    use_chords = np.append(use_chords, False)
                else:
                    use_chords = np.append(use_chords, True)
        if len(fm[use_chords]) < 2:
            raise ValueError('The number of fitted chords should be higher than two')
        out = self.__linear_fit_error(x=fm[use_chords], y=gm[use_chords], sx=dfm[use_chords], sy=dgm[use_chords],
                                      verbose=verbose, use_error=use_error)
        fm_fit = np.arange(-delta_plot+np.min([fm.min(), gm.min()]), delta_plot+np.max([fm.max(), gm.max()]),
                           time_resolution*np.absolute(self.vel.value))
        gm_fit = self.__func_line_decalage(out.beta, fm_fit)

        previous_dt = np.array([])
        fm_decalage = np.array([])
        gm_decalage = np.array([])
        time_decalage = np.array([])

        for i, j in enumerate(chord_name):
            previous_dt = np.append(previous_dt, self.chords[j].lightcurve.dt)
            dist_min = np.array([])
            dtt = np.arange(-time_interval, +time_interval, time_resolution)
            for dt in dtt:
                fm_new = fm[i] + dt*vfm[i]
                gm_new = gm[i] + dt*vgm[i]
                dist = np.sqrt((fm_new - fm_fit)**2 + (gm_new - gm_fit)**2)
                dist_min = np.append(dist_min, dist.min())
            print('dt = {:+6.3f} seconds; chord = {}'.format(previous_dt[i] + dtt[dist_min.argmin()], chord_name[i]))
            fm_decalage = np.append(fm_decalage, fm[i] + dtt[dist_min.argmin()]*vfm[i])
            gm_decalage = np.append(gm_decalage, gm[i] + dtt[dist_min.argmin()]*vgm[i])
            time_decalage = np.append(time_decalage, dtt[dist_min.argmin()])

        for i in range(len(chord_name)):
            out_dic[chord_name[i]] = time_decalage[i]
        if plot:
            plt.figure(figsize=(5, 5))
            plt.title('Before', fontsize=15)
            self.chords.plot_chords(color='blue')
            self.chords.plot_chords(color='red', segment='error')
            plt.plot(fm[use_chords], gm[use_chords], linestyle='None', marker='o', color='k')
            plt.plot(fm[np.invert(use_chords)], gm[np.invert(use_chords)], linestyle='None', marker='x', color='r')
            plt.plot(fm_fit, gm_fit, 'k-')
            plt.xlim(-delta_plot + np.min([fm.min(), gm.min()]), delta_plot + np.max([fm.max(), gm.max()]))
            plt.ylim(-delta_plot + np.min([fm.min(), gm.min()]), delta_plot + np.max([fm.max(), gm.max()]))
            plt.show()
            for i, j in enumerate(chord_name):
                self.chords[j].lightcurve.dt = previous_dt[i] + time_decalage[i]
            plt.figure(figsize=(5, 5))
            plt.title('After', fontsize=15)
            self.chords.plot_chords(color='blue')
            self.chords.plot_chords(color='red', segment='error')
            plt.plot(fm_decalage[use_chords], gm_decalage[use_chords], linestyle='None', marker='o', color='k')
            plt.plot(fm_decalage[np.invert(use_chords)], gm_decalage[np.invert(use_chords)], linestyle='None', marker='x',
                     color='r')
            plt.plot(fm_fit, gm_fit, 'k-')
            plt.xlim(-delta_plot + np.min([fm.min(), gm.min()]), delta_plot + np.max([fm.max(), gm.max()]))
            plt.ylim(-delta_plot + np.min([fm.min(), gm.min()]), delta_plot + np.max([fm.max(), gm.max()]))
            plt.show()
            for i, j in enumerate(chord_name):
                self.chords[j].lightcurve.dt = previous_dt[i]
        return out_dic

    def to_file(self):
        """Saves the occultation data to a file.

        Three files are saved containing the positions and velocities for the
        observations. They are for the positive, negative and error bars positions.

        The format of the files are: positions in f and g, velocities in f and
        g, the Julian Date of the observation, light curve name of the
        corresponding position.
        """
        pos = []
        neg = []
        err = []
        for name, chord in self.chords.items():
            status = chord.status()
            l_name = name.replace(' ', '_')
            if status == 'positive':
                im = chord.lightcurve.immersion
                ime = chord.lightcurve.immersion_err
                f, g, vf, vg = chord.get_fg(time=im, vel=True)
                f1, g1 = chord.get_fg(time=im-ime*u.s)
                f2, g2 = chord.get_fg(time=im+ime*u.s)
                pos.append([f, g, vf, vg, im.jd, l_name+'_immersion'])
                err.append([f1, g1, vf, vg, (im-ime*u.s).jd, l_name+'_immersion_err-'])
                err.append([f2, g2, vf, vg, (im+ime*u.s).jd, l_name+'_immersion_err+'])

                em = chord.lightcurve.emersion
                eme = chord.lightcurve.emersion_err
                f, g, vf, vg = chord.get_fg(time=em, vel=True)
                f1, g1 = chord.get_fg(time=em-eme*u.s)
                f2, g2 = chord.get_fg(time=em+eme*u.s)
                pos.append([f, g, vf, vg, em.jd, l_name+'_emersion'])
                err.append([f1, g1, vf, vg, (em-eme*u.s).jd, l_name+'_emersion_err-'])
                err.append([f2, g2, vf, vg, (em+eme*u.s).jd, l_name+'_emersion_err+'])

            if status == 'negative':
                ini = chord.lightcurve.initial_time
                f, g, vf, vg = chord.get_fg(time=ini, vel=True)
                neg.append([f, g, vf, vg, ini.jd, l_name+'_start'])

                end = chord.lightcurve.end_time
                f, g, vf, vg = chord.get_fg(time=end, vel=True)
                neg.append([f, g, vf, vg, end.jd, l_name+'_end'])
        if len(pos) > 0:
            f = open('occ_{}_pos.txt'.format(self.body.shortname.replace(' ', '_')), 'w')
            for line in pos:
                f.write('{:10.3f} {:10.3f} {:-6.2f} {:-6.2f} {:16.8f} {}\n'.format(*line))
            f.close()
            f = open('occ_{}_err.txt'.format(self.body.shortname.replace(' ', '_')), 'w')
            for line in err:
                f.write('{:10.3f} {:10.3f} {:-6.2f} {:-6.2f} {:16.8f} {}\n'.format(*line))
            f.close()
        if len(neg) > 0:
            f = open('occ_{}_neg.txt'.format(self.body.shortname.replace(' ', '_')), 'w')
            for line in neg:
                f.write('{:10.3f} {:10.3f} {:-6.2f} {:-6.2f} {:16.8f} {}\n'.format(*line))
            f.close()

    def __str__(self):
        """ String representation of the Occultation class
        """
        out = ('Stellar occultation of star {} {} by {}.\n\n'
               'Geocentric Closest Approach: {:.3f}\n'
               'Instant of CA: {}\n'
               'Position Angle: {:.2f}\n'
               'Geocentric shadow velocity: {:.2f}\n'
               'Sun-Geocenter-Target angle:  {:.2f} deg\n'
               'Moon-Geocenter-Target angle: {:.2f} deg\n\n\n'.format(
            self.star._catalogue, self.star.code, self.body.name, self.ca, self.tca.iso,
            self.pa, self.vel, self.predict['S-G-T'].data[0], self.predict['M-G-T'].data[0])
        )

        count = {'positive': 0, 'negative': 0, 'visual': 0}
        string = {'positive': '', 'negative': '', 'visual': ''}
        if len(self.chords) > 0:
            for chord in self.chords.values():
                status = chord.status()
                string[status] += chord.__str__() + '\n'
                count[status] += 1

        if len(self.chords) == 0:
            out += 'No observations reported'
        else:
            out += '\n'.join(['{} {} observations'.format(count[k], k) for k in string.keys() if count[k] > 0])

        out += '\n\n'

        out += '#'*79 + '\n{:^79s}\n'.format('STAR') + '#'*79 + '\n'
        out += self.star.__str__() + '\n\n'

        coord = self.star.get_position(self.tca, observer=self._reference_center)
        try:
            error_star = self.star.error_at(self.tca)
        except:
            error_star = [0, 0]*u.mas
        out += 'Geocentric star coordinate at occultation Epoch ({}):\n'.format(self.tca.iso)
        out += 'RA={} +/- {:.4f}, DEC={} +/- {:.4f}\n\n'.format(
            coord.ra.to_string(u.hourangle, sep='hms', precision=5), error_star[0],
            coord.dec.to_string(u.deg, sep='dms', precision=4), error_star[1])

        out += self.body.__str__() + '\n'

        for status in string.keys():
            if count[status] == 0:
                continue
            out += ('#'*79 + '\n{:^79s}\n'.format(status.upper() + ' OBSERVATIONS') + '#'*79)
            out += '\n\n'
            out += string[status]

        if hasattr(self, 'fitted_params'):
            out += '#'*79 + '\n{:^79s}\n'.format('RESULTS') + '#'*79 + '\n\n'
            out += 'Fitted Ellipse:\n'
            out += '\n'.join(['{}: {:.3f} +/- {:.3f}'.format(k, *self.fitted_params[k])
                              for k in self.fitted_params.keys()]) + '\n'
            polar_radius = self.fitted_params['equatorial_radius'][0]*(1.0-self.fitted_params['oblateness'][0])
            equivalent_radius = np.sqrt(self.fitted_params['equatorial_radius'][0]*polar_radius)
            out += 'polar_radius: {:.3f} km \n'.format(polar_radius)
            out += 'equivalent_radius: {:.3f} km \n'.format(equivalent_radius)
            if not np.isnan(self.body.H):
                H_sun = -26.74
                geometric_albedo = (10**(0.4*(H_sun - self.body.H.value))) * ((u.au.to('km')/equivalent_radius)**2)
                out += 'geometric albedo (V): {:.3f} ({:.1%}) \n'.format(geometric_albedo, geometric_albedo)
            else:
                out += 'geometric albedo (V): not calculated, absolute magnitude (H) is unknown \n'
            out += '\nMinimum chi-square: {:.3f}\n'.format(self.chi2_params['chi2_min'])
            out += 'Number of fitted points: {}\n'.format(self.chi2_params['npts'])
            out += 'Number of fitted parameters: {}\n'.format(self.chi2_params['nparam'])
            out += 'Minimum chi-square per degree of freedom: {:.3f}\n'.format(
                self.chi2_params['chi2_min']/(self.chi2_params['npts'] - self.chi2_params['nparam']))
            out += 'Radial dispersion: {:.3f} +/- {:.3f} km\n'.format(
                self.chi2_params['radial_dispersion'].mean(), self.chi2_params['radial_dispersion'].std(ddof=1))
            out += 'Radial error:      {:.3f} +/- {:.3f} km\n'.format(
                self.chi2_params['radial_error'].mean(), self.chi2_params['radial_error'].std(ddof=1))

            out += '\n' + self.new_astrometric_position(verbose=False)

        return out

    def __func_line_decalage(self, p, x):
        """ Private function returns a linear function, intended for fitting inside the self.check_decalage().
        """
        a, b = p
        return a*x + b

    def __linear_fit_error(self, x, y, sx, sy, verbose=False, use_error=True):
        """ Private function returns a linear fit, intended for fitting inside the self.check_decalage().
        """
        import scipy.odr as odr

        model = odr.Model(self.__func_line_decalage)
        if use_error:
            data = odr.RealData(x=x, y=y, sx=sx, sy=sy)
        else:
            data = odr.RealData(x=x, y=y)
        fit = odr.ODR(data, model, beta0=[0., 1.])
        out = fit.run()
        if verbose:
            print('Linear fit procedure')
            out.pprint()
            print('\n')
        return out



import astropy.units as u
import numpy as np
from astropy.time import Time


# noinspection PyUnresolvedReferences
class MetaStar:
    """Docstring for MetaStar"""

    @property
    def ra(self):
        """Return the Right Ascension of the star
        """
        return self._attributes['RA']

    @ra.setter
    def ra(self, value):
        """Defines the Right Ascension.

        Parameters
        ----------
        value : `int`, `float`, `astropy.unit`
            RA, in deg.
        """
        from astropy.coordinates import Longitude
        self._attributes['RA'] = Longitude(value, unit=u.hourangle)

    @property
    def dec(self):
        """Return the Declination of the star
        """
        return self._attributes['DEC']

    @dec.setter
    def dec(self, value):
        """Defines the Declination.

        Parameters
        ----------
        value : `int`, `float`, `astropy.unit`
            DEC, in deg.
        """
        from astropy.coordinates import Latitude
        self._attributes['DEC'] = Latitude(value, unit=u.deg)

    @property
    def parallax(self):
        """Return the Parallax of the star
        """
        if self.bjones:
            return self._attributes['bjones_par']
        else:
            return self._attributes.get('PAR', 0*u.mas)

    @parallax.setter
    def parallax(self, value):
        """Defines the Parallax.

        Parameters
        ----------
        value : `int`, `float`, `astropy.unit`
            Parallax, in mas.
        """
        par = u.Quantity(value, unit=u.mas)
        if par <= 0*u.mas:
            par = 0*u.mas
        self._attributes['PAR'] = par

    @property
    def distance(self):
        """Return the Distance of the star
        """
        from astropy.coordinates import Distance
        if self.parallax > 0*u.mas:
            return Distance(parallax=self.parallax, allow_negative=False)
        else:
            raise ValueError('SORA is not able to determine distance from paralax {}'.format(self.parallax))

    @property
    def coord(self):
        from astropy.coordinates import SkyCoord
        try:
            return SkyCoord(self.ra, self.dec, self.distance)
        except ValueError:
            return SkyCoord(self.ra, self.dec)

    @property
    def pmra(self):
        """Return the Proper Motion in Right Ascension of the star
        """
        return self._attributes.get('PMRA', 0*u.mas/u.year)

    @pmra.setter
    def pmra(self, value):
        """Defines the Proper Motion in RA*.

        Parameters
        ----------
        value : `int`, `float`, `astropy.unit`
            RA Proper Motion, in mas/year.
        """
        self._attributes['PMRA'] = u.Quantity(value, unit=u.mas/u.year)

    @property
    def pmdec(self):
        """ Return the Proper Motion in Declination of the star
        """
        return self._attributes.get('PMDEC', 0*u.mas/u.year)

    @pmdec.setter
    def pmdec(self, value):
        """Defines the Proper Motion in DEC.

        Parameters
        ----------
        value : `int`, `float`, `astropy.unit`
            DEC Proper Motion, in mas/year.
        """
        self._attributes['PMDEC'] = u.Quantity(value, unit=u.mas/u.year)

    @property
    def rad_vel(self):
        """Return the Radial Velocity of the star
        """
        return self._attributes.get('RAD_VEL', 0*u.km/u.s)

    @rad_vel.setter
    def rad_vel(self, value):
        """Defines the Radial Velocity.

        Parameters
        ----------
        value : `int`, `float`, `astropy.unit`
            Radial Velocity, in km/s.
        """
        self._attributes['RAD_VEL'] = u.Quantity(value, unit=u.km/u.s)

    @property
    def epoch(self):
        """Return the Epoch of the position of the star
        """
        return self._attributes['EPOCH']

    @epoch.setter
    def epoch(self, value):
        """Defines the Epoch of the position.

        Parameters
        ----------
        value : `int`, `float`, `astropy.unit`
            Radial Velocity, in km/s.
        """
        self._attributes['EPOCH'] = Time(value)

    @property
    def bjones(self):
        return self._bjones

    @bjones.setter
    def bjones(self, value):
        from .utils import search_star, choice_star
        if value not in [True, False]:
            raise AttributeError('bjones attribute must be True or False')
        if value and 'bjones_par' not in self._attributes:
            if hasattr(self, 'code'):
                catalogue = search_star(code=self.code, columns=['**'], catalog='I/347/gaia2dis', verbose=self._verbose)
            else:
                catalogue = search_star(coord=self.coord, columns=['**'], radius=2*u.arcsec,
                                        catalog='I/347/gaia2dis', verbose=self._verbose)
            if len(catalogue) == 0:
                raise ValueError('No star was found in the Bailer-Jones catalogue. It does not exist or VizieR is out.')
            catalogue = catalogue[0]
            if len(catalogue) > 1:
                print('{} stars were found within 2 arcsec from given coordinate.'.format(len(catalogue)))
                print('The list below is sorted by distance. Please select the correct star')
                if hasattr(self.mag, 'G'):
                    print('Star G mag: {}'.format(self.mag['G']))
                catalogue = choice_star(catalogue, self.coord, ['RA_ICRS', 'DE_ICRS', 'Source'], source='bjones')
                if catalogue is None:
                    return
            self._attributes['bjones_par'] = ((1.0/catalogue['rest'][0])*u.arcsec).to(u.mas)
            self.meta_bjones = {c: catalogue[c][0] for c in catalogue.columns}
        self._bjones = value

    @property
    def diameter_gaia(self):
        try:
            rad = self.meta_gaia.get('Rad')
            if rad is not None and not np.ma.core.is_masked(rad):
                return 2*np.arctan((rad*u.solRad)/self.distance).to(u.mas)
        except:
            return None



import astropy.units as u
import numpy as np

__all__ = ['draw_ellipse']


def draw_ellipse(equatorial_radius, oblateness=0.0, center_f=0.0, center_g=0.0,
                 position_angle=0.0, center_dot=False, ax=None, **kwargs):
    """Plots an ellipse with the given input parameters.

    Parameters
    ----------
    equatorial_radius : `float`, `int`, default=0
        Semi-major axis of the ellipse.

    oblateness : `float`, `int`, default=0
        Oblateness of the ellipse.

    center_f : `float`, `int`, default=0
        Coordinate of the ellipse (abscissa).

    center_g : `float`, `int`, default=0
        Coordinate of the ellipse (ordinate).

    center_dot : `bool`, default=False
        If True, plots a dot at the center of the ellipse.

    position_angle : `float`, `int`, default= 0
        Pole position angle (Default=0.0).

    ax : `maptlotlib.pyplot.Axes`
        Axis where to plot ellipse.

    **kwargs
        All other parameters. They will be parsed directly by matplotlib.
    """
    import matplotlib.pyplot as plt

    equatorial_radius = np.array(equatorial_radius, ndmin=1)
    oblateness = np.array(oblateness, ndmin=1)
    center_f = np.array(center_f, ndmin=1)
    center_g = np.array(center_g, ndmin=1)
    position_angle = np.array(position_angle, ndmin=1)

    theta = np.linspace(-np.pi, np.pi, 1800)

    ax = ax or plt.gca()

    if len(equatorial_radius) == 1:
        if 'color' not in kwargs:
            kwargs['color'] = 'black'
        if 'lw' not in kwargs:
            kwargs['lw'] = 2
    else:
        if 'color' not in kwargs:
            kwargs['color'] = 'gray'
        if 'lw' not in kwargs:
            kwargs['lw'] = 0.1
        if 'alpha' not in kwargs:
            kwargs['alpha'] = 0.1
        if 'zorder' not in kwargs:
            kwargs['zorder'] = 0.5
    for i in np.arange(len(equatorial_radius)):
        circle_x = equatorial_radius[i] * np.cos(theta)
        circle_y = equatorial_radius[i] * (1.0 - oblateness[i]) * np.sin(theta)
        pos_ang = position_angle[i] * u.deg
        ax.plot(+circle_x * np.cos(pos_ang) + circle_y * np.sin(pos_ang) + center_f[i],
                -circle_x * np.sin(pos_ang) + circle_y * np.cos(pos_ang) + center_g[i],
                **kwargs)
    if center_dot:
        kwargs.pop('lw')
        plt.plot(center_f, center_g, '.', **kwargs)
    plt.axis('equal')



import numpy as np

__all__ = ['ChiSquare']


class ChiSquare:
    """Stores the arrays for all inputs and given chi-square.

    Parameters
    ----------
    chi2 : `array`
        Array with all the chi-square values.

    npts : `int`
        Number of points used in the fit.

    **kwargs
        Any other given input must be an array with the same size as chi2. The
        keyword `name` will be associated as the variable `name` of the given data.

    Example
    -------
    >>> chisquare = ChiSquare(chi2, immersion=t1, emersion=t2)

    ``t1`` and ``t2`` must be an array with the same size as chi2.

    The data can be accessed as

    >>> chisquare.data['immersion']

    """

    def __init__(self, chi2, npts, **kwargs):

        self._names = ['chi2']
        self.data = {'chi2': chi2}
        data_size = len(chi2)
        self.npts = npts
        nparam = 0
        for item in kwargs.keys():
            if kwargs[item].var() != 0:
                nparam += 1
            if len(kwargs[item]) != data_size:
                raise ValueError('{} size must have the same size as given chi2'.format(item))
            self._names.append(item)
            self.data[item] = kwargs[item]
        self.nparam = nparam

    def get_nsigma(self, sigma=1, key=None):
        """Determines the interval of the chi-square within the nth sigma.

        Parameters
        ----------
        sigma : `float`, `int`
            Value of sigma to calculate.

        key : `str`, default=None
            keyword the user desire to obtain results.

        Returns
        -------
        dict
            Dictionary with the average n-sigma value and bondaries.

        Note
        ----
        If a key value is given the mean value within the n-sigma and the error
        bar within the n-sigma are returned.

        If no key is given, a dictionary with: the minimum chi-square, the sigma
        required, the number of points where :math:`chi2 < (chi2_min + sigma^2)`,
        and the mean values and errors for all keys is returned.
        """
        values = np.where(self.data['chi2'] < self.data['chi2'].min() + sigma ** 2)[0]
        output = {'chi2_min': self.data['chi2'].min(), 'sigma': sigma, 'n_points': len(values)}
        for name in self._names[1:]:
            vmax = self.data[name][values].max()
            vmin = self.data[name][values].min()
            output[name] = [(vmax + vmin) / 2.0, (vmax - vmin) / 2.0]
        if key is not None:
            if key not in self._names[1:]:
                raise ValueError('{} is not one of the available keys. Please choose one of {}'
                                 .format(key, self._names[1:]))
            return output[key]
        return output

    def plot_chi2(self, key=None, ax=None):
        """Plots an ellipse using the input parameters.

        Parameters
        ----------
        key : `str`
            Key (parameter) for which to plot the chi squares. If no key  is
            given, it will plot for all parameters.

        ax : `maplotlib.pyplot.Axes`
            Matplotlib plot axes. If none is given, it uses the matplotlib pool
            to identify.
        """
        import matplotlib.pyplot as plt

        sigma_1 = self.get_nsigma(sigma=1)
        sigma_3 = self.get_nsigma(sigma=3)
        if key is not None and (key not in self._names[1:]):
            raise ValueError('{} is not one of the available keys. Please choose one of {}'
                             .format(key, self.data.keys()[1:]))
        k = np.where(self.data['chi2'] < sigma_1['chi2_min'] + 11)
        for name in self._names[1:]:
            if (key is not None) and (key != name):
                continue
            if key is None or not ax:
                ax = plt.gca()
            ax.plot(self.data[name][k], self.data['chi2'][k], 'k.')
            ax.set_ylim(sigma_1['chi2_min'] - 1, sigma_1['chi2_min'] + 10)
            delta = sigma_3[name][1]
            if delta == 0.0:
                delta = 1.0
            ax.set_xlim(sigma_3[name][0] - 3 * delta, sigma_3[name][0] + 3 * delta)
            ax.axhline(sigma_1['chi2_min'] + 1, linestyle='--', color='red')
            ax.axhline(sigma_3['chi2_min'] + 9, linestyle='--', color='red')
            ax.set_xlabel(name, fontsize=20)
            ax.set_ylabel('Chi2', fontsize=20)
            if key is None:
                plt.show()

    def to_file(self, namefile):
        """Saves the data to a file.

        Parameters
        ----------
        namefile : `str`
            Filename to save the data.
        """
        data = np.vstack(([self.data[i] for i in self._names]))
        np.savetxt(namefile, data.T, fmt='%11.5f')
        f = open(namefile + '.label', 'w')
        for i, name in enumerate(self._names):
            f.write('Column {}: {}\n'.format(i + 1, name))
        f.close()

    def get_values(self, sigma=0.0, key=None):
        """Returns all values where the chi-square is within the nth sigma.

        Parameters
        ----------
        sigma : `float`, `int`
            Value of sigma to cut values.

        key : `str`
            Keyword (parameter) that the user desires to obtain results of.

        Returns
        -------
        list or dict : `list`, `dict`
            List or dictionary with chi-square values within the nth sigma, the
            average n-sigma value, and bondaries.

        Note
        ----
        If a `key` is given, it returns a `list` with all the values that are within
        the chosen n-sigma.

        If no `key` is given, it returns a dictionary with the list with all the
        values that are within the n-sigma for all keys.

        If ``sigma=0``, it returns the parameters for the minimum chi-square
        instead of a list.
        """
        values = {}
        if sigma == 0.0:
            k = np.argsort(self.data['chi2'])[0]
        else:
            k = np.where(self.data['chi2'] < self.data['chi2'].min() + sigma ** 2)[0]
        for name in self._names[1:]:
            values[name] = self.data[name][k]
        values = values.get(key, values)
        return values

    def __len__(self):
        return len(self.data['chi2'])

    def __add__(self, other):
        if not isinstance(other, ChiSquare):
            raise TypeError(
                f"unsupported operand type(s) for +: '{self.__class__.__name__}' and '{other.__class.__name}'")
        if self._names != other._names:
            raise ValueError(f"ChiSquare objects does not have the same keys: '{self._names}' and '{other._names}'")
        if self.npts != other.npts:
            raise ValueError(
                f"The number of fitted points are different between the objects: '{self.npts}' and '{other.npts}'")
        params = {key: np.hstack((self.data[key], other.data[key])) for key in self._names}
        chi2 = params.pop('chi2')
        return ChiSquare(chi2=chi2, npts=int((self.npts + other.npts) / 2), **params)

    def __str__(self):
        """ String representation of the ChiSquare Object
        """
        sigma_1 = self.get_nsigma(sigma=1)
        sigma_3 = self.get_nsigma(sigma=3)
        output = ('Minimum chi-square: {:.3f}\n'
                  'Number of fitted points: {}\n'
                  'Number of fitted parameters: {}\n'
                  'Minimum chi-square per degree of freedom: {:.3f}\n'.format(
                      sigma_1['chi2_min'], self.npts, self.nparam,
                      sigma_1['chi2_min']/(self.npts-self.nparam))
                  )
        for name in self._names[1:]:
            output += ('\n{}:\n'
                       '    1-sigma: {:.3f} +/- {:.3f}\n'
                       '    3-sigma: {:.3f} +/- {:.3f}\n'.format(
                           name, *sigma_1[name], *sigma_3[name])
                       )
        return output



import functools
import warnings

next_major_version = 'v1.0'

warnings.simplefilter('always', FutureWarning)


# This decorator gets an argument that is being deprecated
def deprecated_alias(**aliases):
    def deco(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            rename_kwargs(f.__name__, kwargs, aliases)
            return f(*args, **kwargs)

        return wrapper

    return deco


def rename_kwargs(func_name, kwargs, aliases):
    for alias, new in aliases.items():
        if alias in kwargs:
            if new in kwargs:
                raise TypeError('{} received both {} and {}'.format(
                    func_name, alias, new))
            warnings.warn("'{}' is deprecated and will be removed in {}; please use '{}'".
                          format(alias, next_major_version, new), FutureWarning)
            kwargs[new] = kwargs.pop(alias)


def deprecated_function(message):
    def deco(func):
        """This is a decorator which can be used to mark functions
        as deprecated. It will result in a warning being emitted
        when the function is used."""

        @functools.wraps(func)
        def new_func(*args, **kwargs):
            warnings.warn("{} is deprecated and will be removed in {}; {}".
                          format(func.__name__, next_major_version, message),
                          category=FutureWarning,
                          stacklevel=2)
            return func(*args, **kwargs)

        return new_func

    return deco



import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord

from sora.config import input_tests
from sora.config.decorators import deprecated_alias

__all__ = ['van_belle', 'kervella']


@deprecated_alias(log='verbose')  # remove this line for v1.0
def search_star(**kwargs):
    """Searches position on VizieR and returns a catalogue.

    Parameters
    ----------
    coord : `str`, `astropy.coordinates.SkyCoord`
        Coordinate to perform the search.

    code : `str`
        Gaia Source_id of the star.

    columns : `list`
        List of strings with the name of the columns to retrieve.

    radius : `int`, `float`, `astropy.unit.quantity`
        Radius to search around coordinates.

    catalog : `str`
        VizieR catalogue to search.


    Returns
    -------
    catalogue : `astropy.Table`
        An astropy Table with the catalogue information.
    """
    from astroquery.vizier import Vizier

    input_tests.check_kwargs(kwargs, allowed_kwargs=['catalog', 'code', 'columns', 'coord', 'verbose', 'radius'])
    row_limit = 100
    if 'verbose' in kwargs and kwargs['verbose']:
        print('\nDownloading star parameters from {}'.format(kwargs['catalog']))
    vquery = Vizier(columns=kwargs['columns'], row_limit=row_limit, timeout=600)
    if 'code' in kwargs:
        catalogue = vquery.query_constraints(catalog=kwargs['catalog'], Source=kwargs['code'], cache=False)
    elif 'coord' in kwargs:
        catalogue = vquery.query_region(kwargs['coord'], radius=kwargs['radius'], catalog=kwargs['catalog'], cache=False)
    else:
        raise ValueError('At least a code or coord should be given as input')
    return catalogue


def van_belle(magB=None, magV=None, magK=None):
    """Determines the diameter of a star in mas using equations from van Belle (1999).

    See: Publi. Astron. Soc. Pacific 111, 1515-1523:.

    Parameters
    ----------
    magB : `float`, default=None
        The magnitude B of the star.

    magV : `float`, default=None
        The magnitude V of the star.

    magK : `float`, default=None
        The magnitude K of the star.


    Note
    ----
    If any of those values is 'None', 'nan' or higher than 49, it is not considered.
    """
    if magB is None or np.isnan(magB) or magB > 49:
        magB = np.nan
    if magV is None or np.isnan(magV) or magV > 49:
        magV = np.nan
    if magK is None or np.isnan(magK) or magK > 49:
        magK = np.nan

    def calc_diameter(a1, a2, mag):
        return 10**(a1 + a2*(mag - magK) - 0.2*mag)

    params = {'sg': {'B': [0.648, 0.220], 'V': [0.669, 0.223]},
              'ms': {'B': [0.500, 0.290], 'V': [0.500, 0.264]},
              'vs': {'B': [0.840, 0.211], 'V': [0.789, 0.218]}}

    mag = np.array([magB, magV])
    diameter = {}
    for st in ['sg', 'ms', 'vs']:
        diameter_s = {}
        for i, m in enumerate(['B', 'V']):
            diam = calc_diameter(*params[st][m], mag[i])
            if not np.isnan(diam):
                diameter_s[m] = calc_diameter(*params[st][m], mag[i])*u.mas
        if diameter_s:
            diameter[st] = diameter_s
    return diameter


def kervella(magB=None, magV=None, magK=None):
    """Determines the diameter of a star in mas using equations from Kervella et. al (2004).

    See: A&A Vol. 426, No.  1:.

    Parameters
    ----------
    magB: `float`, default=None
        The magnitude B of the star.

    magV: `float`, default=None
        The magnitude V of the star.

    magK: `float`, default=None
        The magnitudes K of the star.


    Note
    ----
    If any of those values is 'None', 'nan' or higher than 49, it is not considered.

    """
    if magB is None or np.isnan(magB) or magB > 49:
        magB = np.nan
    if magV is None or np.isnan(magV) or magV > 49:
        magV = np.nan
    if magK is None or np.isnan(magK) or magK > 49:
        magK = np.nan
    const1 = np.array([0.0755, 0.0535])
    const2 = np.array([0.5170, 0.5159])
    mag = np.array([magV, magB])
    vals = 10**(const1*(mag-magK)+const2-0.2*magK)
    diameter = {}
    if not np.isnan(vals[0]):
        diameter['V'] = vals[0]*u.mas
    if not np.isnan(vals[1]):
        diameter['B'] = vals[1]*u.mas
    return diameter


def spatial_motion(ra, dec, pmra, pmdec, parallax=0, rad_vel=0,  dt=0, cov_matrix=None):
    """Applies spatial motion to star coordinate.

    Parameters
    ----------
    ra `int`, `float`
        Right Ascension of the star at t=0 epoch, in deg.

    dec : `int`, `float`
        Declination of the star at t=0 epoch, in deg.

    pmra : `int`, `float`
        Proper Motion in RA of the star at t=0 epoch, in mas/year.

    pmdec : `int`, `float`
        Proper Motion in DEC of the star at t=0 epoch, in mas/year.

    parallax : `int`, `float`
        Parallax of the star at t=0 epoch, in mas.

    rad_vel : `int`, `float`
        Radial Velocity of the star at t=0 epoch, in km/s.

    dt : `int`, `float`
        Variation of time from catalogue epoch, in days.

    cov_matrix : `2D-array`
        6x6 covariance matrix.
    """
    import astropy.constants as const

    A = (1*u.AU).to(u.km).value  # Astronomical units in km
    c = const.c.to(u.km/u.year).value  # light velocity

    par = True
    # Eliminate negative or zero parallaxes
    if parallax is None or parallax <= 0:
        par = False
        parallax = 1e-4

    if cov_matrix is not None and cov_matrix.shape != (6, 6):
        raise ValueError('Covariance matrix must be a 6x6 matrix')

    ra0 = u.Quantity(ra, unit=u.deg).to(u.rad).value
    dec0 = u.Quantity(dec, unit=u.deg).to(u.rad).value
    parallax0 = u.Quantity(parallax, unit=u.mas).to(u.rad).value
    pmra0 = u.Quantity(pmra, unit=u.mas/u.year).to(u.rad/u.year).value
    pmdec0 = u.Quantity(pmdec, unit=u.mas/u.year).to(u.rad/u.year).value
    rad_vel0 = u.Quantity(rad_vel, unit=u.km/u.s).to(u.AU/u.year).value
    dt = u.Quantity(dt, unit=u.day).to(u.year).value

    # normal triad relative to the celestial sphere
    # p0 points to growing RA, q0 to growing DEC and r0 to growing distance.
    p0 = np.array([-np.sin(ra0), np.cos(ra0), 0.0]).T
    q0 = np.array([-np.sin(dec0)*np.cos(ra0), -np.sin(dec0)*np.sin(ra0), np.cos(dec0)])
    r0 = np.array([np.cos(dec0)*np.cos(ra0), np.cos(dec0)*np.sin(ra0), np.sin(dec0)])

    b0 = A/parallax0
    tau_0 = b0/c
    tau_A = A/c

    vec_b0 = b0*r0  # distance vector
    vec_u0 = vec_b0/np.linalg.norm(vec_b0)
    vec_mi0 = np.array(p0*pmra0 + q0*pmdec0)  # proper motion vector

    mi_r0 = rad_vel0/b0
    mi0 = np.sqrt(pmra0**2+pmdec0**2)  # total proper motion

    v0 = b0*(r0*mi_r0+vec_mi0)  # apparent space velocity
    v_r0 = np.linalg.norm(v0)

    # Scaling factors of time, distance and velocity due to light time
    f_T = ((dt + 2*tau_0)/(tau_0+(1-v_r0/c)*dt + np.sqrt(np.linalg.norm((vec_b0+v0*dt))**2
           + (2*dt/(c**2*tau_0))*np.linalg.norm(np.cross(v0, vec_b0))**2)/c))
    f_D = np.sqrt(1+2*mi_r0*dt*f_T + (mi0**2 + mi_r0**2)*(dt*f_T)**2)
    f_V = (1 + (tau_A/parallax0)*(mi_r0*(f_D - 1) + f_D*(mi0**2 + mi_r0**2)*dt*f_T))

    vec_u = (r0*(1 + mi_r0*dt*f_T) + vec_mi0*dt*f_T)*f_D
    vec_mi = (vec_mi0*(1 + mi_r0*dt*f_T) - vec_u0*mi0**2*f_T)*f_D**3*f_V
    mi_r = (mi_r0 + (mi0**2 + mi_r0**2)*dt*f_T)*f_D**2*f_V

    dec = np.arcsin(vec_u[2])  # new dec
    ra = np.arctan2(vec_u[1]/np.cos(dec), vec_u[0]/np.cos(dec))  # new ra

    parallax = parallax0*f_D  # new parallax
    new_dist = A/parallax  # new distance

    if par:
        coord = SkyCoord(ra*u.rad, dec*u.rad, new_dist*u.km)
    else:
        coord = SkyCoord(ra*u.rad, dec*u.rad)

    if cov_matrix is None:
        return coord

    p = np.array([-np.sin(ra), np.cos(ra), 0.0])
    q = np.array([-np.sin(dec)*np.cos(ra), -np.sin(dec)*np.sin(ra), np.cos(dec)])

    Z = np.sqrt(1 + (dt + 2*tau_A/parallax0)*mi0**2*dt + (2 + mi_r0*dt)*mi_r0*dt)
    Y = parallax0*dt + tau_A*(1 + Z - mi_r0*dt)
    X = parallax0*dt + 2*tau_A

    # partial derivatives of the logarithm of the velocity factor
    # chi_parallax = (1/parallax0)*(1 - f_V)  # not used
    chi_pm = (tau_A/parallax0)*dt*f_T*f_D*(mi_r*dt*f_T - 2*f_V)
    chi_r = (tau_A/parallax0)*(f_V + f_D*(f_V + (1 + mi_r0*dt*f_T)*(mi_r*dt*f_T - 2*f_V)))
    # chi_T = -(tau_A/parallax0)*f_D**3*mi0**2*dt*f_T  # not used

    # partial derivatives of the logarithmic of the time factor
    psi_parallax = dt/X - (dt/Y)*(1 - (mi0**2*tau_A**2)/(Z*parallax0**2))
    psi_pm = -((dt*tau_A)/(Y*Z))*(dt + 2*tau_A/parallax0)
    psi_r = -(dt*tau_A/Y)*((1 + mi_r0*dt)/Z - 1)

    ni = vec_mi*(1 - dt*f_T*(3*mi_r/f_V + (tau_A/parallax0)*mi0**2*f_D**3*f_V)) - vec_mi0*f_D**3*f_V
    eta = mi_r*(1 - dt*f_T*(2*mi_r/f_V + (tau_A/parallax0)*mi0**2*f_D**3*f_V)) - mi_r0*f_D**2*f_V

    p_l = p/np.linalg.norm(p)
    q_l = q/np.linalg.norm(q)

    pmra = np.dot(p_l, vec_mi)  # new pmra
    pmdec = np.dot(q_l, vec_mi)  # new pmdec

    # Jacobian matrix
    J = np.zeros((6, 6))

    # d(alpha)/d(valores inicias)
    J[0, 0] = np.dot(p_l, p0)*(1 + mi_r0*dt*f_T)*f_D - np.dot(p_l, r0)*pmra0*dt*f_T*f_D
    J[0, 1] = np.dot(p_l, q0)*(1 + mi_r0*dt*f_T)*f_D - np.dot(p_l, r0)*pmdec0*dt*f_T*f_D
    J[0, 2] = - np.dot(p_l, r0)*f_D*psi_parallax
    J[0, 3] = np.dot(p_l, p0)*dt*f_T*f_D - np.dot(p_l, r0)*pmra0*f_D*psi_pm
    J[0, 4] = np.dot(p_l, q0)*dt*f_T*f_D - np.dot(p_l, r0)*pmdec0*f_D*psi_pm
    J[0, 5] = - pmra*(dt*f_T)**2/f_V - np.dot(p_l, r0)*f_D*psi_r

    # d(delta)/d(valores inicias)
    J[1, 0] = np.dot(q_l, p0)*(1 + mi_r0*dt*f_T)*f_D - np.dot(q_l, r0)*pmra0*dt*f_T*f_D
    J[1, 1] = np.dot(q_l, q0)*(1 + mi_r0*dt*f_T)*f_D - np.dot(q_l, r0)*pmdec0*dt*f_T*f_D
    J[1, 2] = - np.dot(q_l, r0)*f_D*psi_parallax
    J[1, 3] = np.dot(q_l, p0)*dt*f_T*f_D - np.dot(q_l, r0)*pmra0*f_D*psi_pm
    J[1, 4] = np.dot(q_l, q0)*dt*f_T*f_D - np.dot(q_l, r0)*pmdec0*f_D*psi_pm
    J[1, 5] = - pmdec*(dt*f_T)**2/f_V - np.dot(q_l, r0)*f_D*psi_r

    # d(parallax)/d(valores inicias)
    J[2, 0] = 0
    J[2, 1] = 0
    J[2, 2] = f_D - parallax*(mi_r*dt*f_T/f_V)*psi_parallax
    J[2, 3] = - parallax*pmra0*((dt*f_T)**2*f_D**2 + (mi_r*dt*f_T/f_V)*psi_pm)
    J[2, 4] = - parallax*pmdec0*((dt*f_T)**2*f_D**2 + (mi_r*dt*f_T/f_V)*psi_pm)
    J[2, 5] = - parallax*((1 + mi_r0*dt*f_T)*dt*f_T*f_D**2 + (mi_r*dt*f_T/f_V)*psi_r)

    # d(pmra)/d(valores inicias)
    J[3, 0] = - np.dot(p_l, p0)*mi0**2*dt*f_T*f_D**3*f_V - np.dot(p_l, r0)*pmra0*(1+mi_r0*dt*f_T)*f_D**3*f_V
    J[3, 1] = - np.dot(p_l, q0)*mi0**2*dt*f_T*f_D**3*f_V - np.dot(p_l, r0)*pmdec0*(1+mi_r0*dt*f_T)*f_D**3*f_V
    J[3, 2] = np.dot(p_l, ni)*psi_parallax
    J[3, 3] = np.dot(p_l, p0)*(1 + mi_r0*dt*f_T)*f_D**3*f_V - 2*np.dot(p_l, r0)*pmra0*dt*f_T*f_D**3*f_V \
        - 3*pmra*pmra0*(dt*f_T)**2*f_D**2*f_V + pmra*pmra0*chi_pm + np.dot(p_l, ni)*pmra0*psi_pm
    J[3, 4] = np.dot(p_l, q0)*(1 + mi_r0*dt*f_T)*f_D**3*f_V - 2*np.dot(p_l, r0)*pmdec0*dt*f_T*f_D**3*f_V \
        - 3*pmra*pmdec0*(dt*f_T)**2*f_D**2*f_V + pmra*pmdec0*chi_pm + np.dot(p_l, ni)*pmdec0*psi_pm
    J[3, 5] = np.dot(p_l, (vec_mi0*f_D - 3*vec_mi*(1 + mi_r0*dt*f_T)))*dt*f_T*f_D**2*f_V

    # d(pmdec)/d(valores inicias)
    J[4, 0] = - np.dot(q_l, p0)*mi0**2*dt*f_T*f_D**3*f_V - np.dot(q_l, r0)*pmra0*(1+mi_r0*dt*f_T)*f_D**3*f_V
    J[4, 1] = - np.dot(q_l, q0)*mi0**2*dt*f_T*f_D**3*f_V - np.dot(q_l, r0)*pmdec0*(1+mi_r0*dt*f_T)*f_D**3*f_V
    J[4, 2] = np.dot(q_l, ni)*psi_parallax
    J[4, 3] = np.dot(q_l, p0)*(1 + mi_r0*dt*f_T)*f_D**3*f_V - 2*np.dot(q_l, r0)*pmra0*dt*f_T*f_D**3*f_V \
        - 3*pmdec*pmra0*(dt*f_T)**2*f_D**2*f_V + pmdec*pmra0*chi_pm + np.dot(q_l, ni)*pmra0*psi_pm
    J[4, 4] = np.dot(q_l, q0)*(1 + mi_r0*dt*f_T)*f_D**3*f_V - 2*np.dot(q_l, r0)*pmdec0*dt*f_T*f_D**3*f_V \
        - 3*pmdec*pmdec0*(dt*f_T)**2*f_D**2*f_V + pmdec*pmdec0*chi_pm + np.dot(q_l, ni)*pmdec0*psi_pm
    J[4, 5] = np.dot(q_l, (vec_mi0*f_D - 3*vec_mi*(1 + mi_r0*dt*f_T)))*dt*f_T*f_D**2*f_V

    # d(rad_vel)/d(valores inicias)
    J[5, 0] = 0
    J[5, 1] = 0
    J[5, 2] = eta*psi_parallax
    J[5, 3] = 2*pmra0*(1 + mi_r0*dt*f_T)*dt*f_T*f_D**4*f_V + pmra0*mi_r*chi_pm + pmra0*eta*psi_pm
    J[5, 4] = 2*pmdec0*(1 + mi_r0*dt*f_T)*dt*f_T*f_D**4*f_V + pmdec0*mi_r*chi_pm + pmdec0*eta*psi_pm
    J[5, 5] = ((1 + mi_r0*dt*f_T)**2 - mi0**2*(dt*f_T)**2)*f_D**4*f_V + mi_r*chi_r + eta*psi_r

    # Propagated covariance matrix
    C = np.matmul(J, np.matmul(cov_matrix, J.T))

    err = np.array([np.sqrt(C[i, i]) for i in np.arange(3)])
    if not par:
        err = err[:2]

    return coord, err


def choice_star(catalogue, coord, columns, source):
    from astropy.table import Table

    tstars = SkyCoord(catalogue[columns[0]], catalogue[columns[1]])
    sep = tstars.separation(coord)
    k = sep.argsort()
    while True:
        t = Table()
        t['num'] = np.arange(len(tstars))+1
        t['dist(")'] = sep[k].arcsec
        t['dist(")'].format = '6.3f'
        for c in columns[2:]:
            t[c] = catalogue[c][k].quantity.value
            t[c].format = '6.3f'
        t['RA___ICRS___DEC'] = tstars[k].to_string('hmsdms', precision=4)
        t.pprint_all()
        print('  0: Cancel')
        choice = int(input('Choose the corresponding number of the correct star: '))
        if choice in np.arange(len(k)+1):
            break
        print('{} is not a valid choice. Please select the correct star'.format(choice))
    if choice == 0:
        if source == 'gaia':
            raise ValueError('It was not possible to define a star')
        elif source == 'nomad':
            print('No magnitudes were obtained from NOMAD')
            return
        elif source == 'bjones':
            print('It was not possible to define a star')
            return
    return catalogue[[k[choice-1]]]



import sys


def progressbar(it, prefix="", size=40, file=sys.stdout):
    count = len(it)

    def show(j):
        x = int(size*j/count)
        file.write("%s |%s%s|  - %i%% \r" % (prefix, "█"*x, "."*(size-x), 100*j/count))
        file.flush()

    show(0)
    for i, item in enumerate(it):
        yield item
        show(i+1)
    file.write("\n")
    file.flush()


def progressbar_show(j, count, prefix='', size=40, file=sys.stdout):
    x = int(size*j/count)
    file.write("%s |%s%s|  - %i%% \r" % (prefix, "█"*x, "."*(size-x), 100*j/count))
    file.flush()
    if x >= size:
        file.write("\n")
        file.flush()
